from __future__ import annotations
from . import get_connection
import argparse
import logging
from ..discover import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ...display import bcolors
from ... import vault
from ...discovery_common.infrastructure import Infrastructure
from ...discovery_common.record_link import RecordLink
from ...discovery_common.user_service import UserService
from ...discovery_common.jobs import Jobs
from ...discovery_common.constants import (PAM_USER, PAM_DIRECTORY, PAM_MACHINE, PAM_DATABASE, VERTICES_SORT_MAP,
                                        DIS_INFRA_GRAPH_ID, RECORD_LINK_GRAPH_ID, USER_SERVICE_GRAPH_ID,
                                        DIS_JOBS_GRAPH_ID)
from ...discovery_common.types import (DiscoveryObject, DiscoveryUser, DiscoveryDirectory, DiscoveryMachine,
                                    DiscoveryDatabase, JobContent)
from ...discovery_common.dag_sort import sort_infra_vertices
from ...keeper_dag import DAG
from ...keeper_dag.connection.commander import Connection as CommanderConnection
from ...keeper_dag.connection.local import Connection as LocalConnection
from ...keeper_dag.vertex import DAGVertex
from ...keeper_dag.edge import DAGEdge
from typing import Optional, Union, TYPE_CHECKING

Connection = Union[CommanderConnection, LocalConnection]
if TYPE_CHECKING:
    from ...vault import TypedRecord
    from ...params import KeeperParams


class PAMDebugGraphCommand(PAMGatewayActionDiscoverCommandBase):

    NO_RECORD = "NO RECORD"
    OTHER = "OTHER"

    parser = argparse.ArgumentParser(prog='pam-action-debug-graph')

    # The record to base everything on.
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name or UID.')
    parser.add_argument('--type', '-t', required=True, choices=['infra', 'rl', 'service', 'jobs'],
                        dest='graph_type', action='store', help='Graph type', default='infra')
    parser.add_argument('--raw', required=False, dest='raw', action='store_true',
                        help='Render raw graph. Will render corrupt graphs.')

    parser.add_argument('--list', required=False, dest='do_text_list', action='store_true',
                        help='List items in a list.')

    parser.add_argument('--render', required=False, dest='do_render', action='store_true',
                        help='Render a graph')
    parser.add_argument('--file', '-f', required=False, dest='filepath', action='store',
                        default="keeper_graph", help='Base name for the graph file.')
    parser.add_argument('--format', required=False, choices=['raw', 'dot', 'twopi', 'patchwork'],
                        dest='format', default="dot", action='store', help='The format of the graph.')
    parser.add_argument('--debug-gs-level', required=False, dest='debug_level', action='store',
                        help='GraphSync debug level. Default is 0', type=int, default=0)

    mapping = {
        PAM_USER: {"order": 1, "sort": "_sort_name", "item": DiscoveryUser, "key": "user"},
        PAM_DIRECTORY: {"order": 1, "sort": "_sort_name", "item": DiscoveryDirectory, "key": "host_port"},
        PAM_MACHINE: {"order": 2, "sort": "_sort_host", "item": DiscoveryMachine, "key": "host"},
        PAM_DATABASE: {"order": 3, "sort": "_sort_host", "item": DiscoveryDatabase, "key": "host_port"},
    }

    graph_id_map = {
        "infra": DIS_INFRA_GRAPH_ID,
        "rl": RECORD_LINK_GRAPH_ID,
        "service": USER_SERVICE_GRAPH_ID,
        "jobs": DIS_JOBS_GRAPH_ID
    }

    def get_parser(self):
        return PAMDebugGraphCommand.parser

    def _do_text_list_infra(self, params: KeeperParams, gateway_context: GatewayContext, debug_level: int = 0,
                            indent: int = 0):

        infra = Infrastructure(record=gateway_context.configuration, params=params, logger=logging,
                               debug_level=debug_level)
        infra.load(sync_point=0)

        try:
            configuration = infra.get_root.has_vertices()[0]
        except (Exception,):
            print(f"{bcolors.FAIL}Could not find the configuration in the infrastructure graph. "
                  f"Has discovery been run for this gateway?{bcolors.ENDC}")

            return

        line_start = {
            0: "",
            1: "* ",
            2: "- ",
        }

        color_func = {
            0: self._h,
            1: self._gr,
            2: self._p,
            3: self._b
        }

        def _handle(current_vertex: DAGVertex, indent: int = 0, last_record_type: Optional[str] = None):

            if current_vertex.active is False:
                return

            pad = ""
            if indent > 0:
                pad = "".ljust(4 * indent, ' ')

            text = ""
            ls = line_start.get(indent, "  ")
            cf = color_func.get(indent, self._p)

            if current_vertex.active is False:
                text += f"{pad}{current_vertex.uid} " + self._f("(Inactive)")
            elif current_vertex.corrupt is False:
                current_content = DiscoveryObject.get_discovery_object(current_vertex)
                if current_content.record_uid is None:
                    text += f"{pad}{ls}{current_vertex.uid}; {current_content.title} does not have a record."
                else:
                    record = vault.KeeperRecord.load(params, current_content.record_uid)  # type: Optional[TypedRecord]
                    if record is not None:
                        text += f"{pad}{ls}" + cf(f"{current_vertex.uid}; {record.title}; {record.record_uid}")
                    else:
                        text += f"{pad}{ls}" + cf(f"{current_vertex.uid}; {current_content.title}; " +
                                                  self._f("have record uid, record does not exists, "
                                                          "might have to sync."))
            else:
                text += f"{pad}{current_vertex.uid} " + self._f("(Corrupt)")

            print(text)

            record_type_to_vertices_map = sort_infra_vertices(current_vertex)
            # Process the record type by their map order in ascending order.

            # Sort the record types by their order in the constant.
            # 'order' is an int.
            for record_type in sorted(record_type_to_vertices_map, key=lambda i: VERTICES_SORT_MAP[i]['order']):
                for vertex in record_type_to_vertices_map[record_type]:
                    if last_record_type is None or last_record_type != record_type:
                        if indent == 0:
                            print(f"{pad}  {self._b(self._n(record_type))}")
                        last_record_type = record_type

                    _handle(vertex, indent=indent+1)

        print("")
        _handle(configuration, indent=indent)
        print("")

    def _do_text_list_rl(self, params: KeeperParams, gateway_context: GatewayContext, debug_level: int = 0,
                         indent: int = 0):

        print("")

        pad = ""
        if indent > 0:
            pad = "".ljust(4 * indent, ' ')

        record_link = RecordLink(record=gateway_context.configuration, params=params, logger=logging,
                                 debug_level=debug_level)
        configuration = record_link.dag.get_root
        
        record = vault.KeeperRecord.load(params, configuration.uid)  # type: Optional[TypedRecord]
        if record is None:
            print(self._f("Configuration record does not exists."))
            return
        
        print(self._h(f"{pad}{record.record_type}, {record.title}, {record.record_uid}"))

        if configuration.has_data is True:
            try:
                data = configuration.content_as_dict
                print(f"{pad}  . data")
                for k, v in data.items():
                    print(f"{pad}    + {k} = {v}")
            except (Exception,):
                print(f"{pad}    ! data not JSON")

        def _group(configuration_vertex: DAGVertex) -> dict:

            group = {
                PAM_USER: [],
                PAM_DIRECTORY: [],
                PAM_DATABASE: [],
                PAM_MACHINE: [],
                PAMDebugGraphCommand.NO_RECORD: [],
                PAMDebugGraphCommand.OTHER: []
            }

            for vertex in configuration_vertex.has_vertices():
                record = vault.KeeperRecord.load(params, vertex.uid)  # type: Optional[TypedRecord]
                if record is None:
                    group[PAMDebugGraphCommand.NO_RECORD].append({
                        "v": vertex
                    })
                    continue
                rt = record.record_type
                if rt not in group:
                    rt = PAMDebugGraphCommand.OTHER
                group[rt].append({
                    "v": vertex,
                    "r": record
                })

            return group
        
        group = _group(configuration)
        
        for record_type in [PAM_USER, PAM_DIRECTORY, PAM_MACHINE, PAM_DATABASE]:
            if len(group[record_type]) > 0:
                print(f"{pad}  " + self._b(self._n(record_type)))
                for item in group[record_type]:
                    vertex = item.get("v")  # type: DAGVertex
                    record = item.get("r")  # type: TypedRecord
                    text = self._gr(f"{record.title}; {record.record_uid}")
                    if vertex.active is False:
                        text += " " + self._f("Inactive")
                    print(f"{pad}    * {text}")

                    # These are cloud users
                    if record_type == PAM_USER:
                        acl = record_link.get_acl(vertex.uid, configuration.uid)
                        if acl is None:
                            print(f"{pad}      {self._f('missing ACL')}")
                        else:
                            if acl.is_admin is True:
                                print(f"{pad}        . is the {self._b('Admin')}")
                            if acl.belongs_to is True:
                                print(f"{pad}      . belongs to this resource")
                            else:
                                print(f"{pad}      . looks like directory user")

                            if acl.rotation_settings:
                                if acl.rotation_settings.noop is True:
                                    print(f"{pad}      . is a NOOP")
                                if acl.rotation_settings.disabled is True:
                                    print(f"{pad}      . rotation is disabled")
                        continue

                    if vertex.has_data is True:
                        try:
                            data = vertex.content_as_dict
                            print(f"{pad}      . data")
                            for k, v in data.items():
                                print(f"{pad}        + {k} = {v}")
                        except (Exception,):
                            print(f"{pad}        ! data not JSON")

                    children = vertex.has_vertices()
                    if len(children) > 0:
                        bad = []
                        for child in children:
                            child_record = vault.KeeperRecord.load(params, child.uid)  # type: Optional[TypedRecord]
                            if child_record is None:
                                if child.active is True:
                                    bad.append(self._f(f"- Record UID {child.uid} does not exists."))
                                continue
                            else:
                                print(f"{pad}      - {child_record.title}; {child_record.record_uid}")
                                acl = record_link.get_acl(child.uid, vertex.uid)
                                if acl is None:
                                    print(f"{pad}        {self._f('missing ACL')}")
                                else:
                                    if acl.is_admin is True:
                                        print(f"{pad}        . is the {self._b('Admin')}")
                                    if acl.belongs_to is True:
                                        print(f"{pad}        . belongs to this resource")
                                    else:
                                        print(f"{pad}        . looks like directory user")

                                if child.has_data is True:
                                    try:
                                        data = child.content_as_dict
                                        print(f"{pad}        . data")
                                        for k, v in data.items():
                                            print(f"{pad}          + {k} = {v}")
                                    except (Exception,):
                                        print(f"{pad}          ! data not JSON")
                        for i in bad:
                            print("{pad}      " + i)

        if len(group[PAMDebugGraphCommand.OTHER]) > 0:
            print(f"{pad}  " + self._b("Other PAM Types"))
            for item in group[PAMDebugGraphCommand.OTHER]:
                vertex = item.get("v")  # type: DAGVertex
                record = item.get("r")  # type: TypedRecord
                text = self._gr(f"{record.record_type}; {record.title}; {record.record_uid}")
                if vertex.active is False:
                    text += " " + self._f("Inactive")
                print(f"{pad}    * {text}")

        if len(group[PAMDebugGraphCommand.NO_RECORD]) > 0:

            # TODO: Check the infra graph for information
            print(f"{pad}  " + self._b(self._n("In Graph, No Vault Record")))
            for item in group[PAMDebugGraphCommand.NO_RECORD]:
                vertex = item.get("v")  # type: DAGVertex
                print(f"{pad}    * {vertex.uid}")


    def _do_text_list_service(self, params: KeeperParams, gateway_context: GatewayContext, debug_level: int = 0,
                              indent: int = 0):

        user_service = UserService(record=gateway_context.configuration, params=params, logger=logging,
                                   debug_level=debug_level)
        configuration = user_service.dag.get_root

        def _handle(current_vertex: DAGVertex, parent_vertex: Optional[DAGVertex] = None, indent: int = 0):

            pad = ""
            if indent > 0:
                pad = "".ljust(2 * indent, ' ') + "* "

            record = vault.KeeperRecord.load(params, current_vertex.uid)  # type: Optional[TypedRecord]
            if record is None:
                if current_vertex.active is False:
                    print(f"{pad}Record {current_vertex.uid} does not exists, inactive in the graph.")
                else:
                    print(f"{pad}Record {current_vertex.uid} does not exists, active in the graph.")
                return
            elif current_vertex.active is False:
                print(f"{pad}{record.record_type}, {record.title}, {record.record_uid} exists, "
                      "inactive in the graph.")
                return

            acl_text = ""
            acl = user_service.get_acl(parent_vertex, current_vertex)
            if acl is not None:
                acl_text = self._f("None")
                acl_parts = []
                if acl.is_service is True:
                    acl_parts.append(self._bl("Service"))
                if acl.is_task is True:
                    acl_parts.append(self._bl("Task"))
                if len(acl_parts) > 0:
                    acl_text = ", ".join(acl_parts)
                acl_text = f"- {acl_text}"

            print(f"{pad}{record.record_type}, {record.title}, {record.record_uid}{acl_text}")

            for vertex in current_vertex.has_vertices():
                _handle(current_vertex=vertex, parent_vertex=current_vertex, indent=indent+1)

        _handle(current_vertex=configuration, parent_vertex=None, indent=indent)

    def _do_text_list_jobs(self, params: KeeperParams, gateway_context: GatewayContext, debug_level: int = 0,
                           indent: int = 0):

        infra = Infrastructure(record=gateway_context.configuration, params=params, logger=logging,
                               debug_level=debug_level, fail_on_corrupt=False)
        infra.load(sync_point=0)

        pad = ""
        if indent > 0:
            pad = "".ljust(2 * indent, ' ') + "* "

        conn = get_connection(params)
        graph_sync = DAG(conn=conn, record=gateway_context.configuration, logger=logging, debug_level=debug_level,
                         graph_id=DIS_JOBS_GRAPH_ID)
        graph_sync.load(0)
        configuration = graph_sync.get_root
        vertices = configuration.has_vertices()
        if len(vertices) == 0:
            print(self._f(f"The jobs graph has not been initialized. Only has root vertex."))
            return

        vertex = vertices[0]
        if vertex.has_data is False:
            print(self._f(f"The job vertex does not contain any data"))
            return

        current_json = vertex.content_as_str
        if current_json is None:
            print(self._f(f"The current job vertex content is None"))
            return

        content = JobContent.model_validate_json(current_json)
        print(f"{pad}{self._b('Active Job ID')}: {content.active_job_id}")
        print("")
        print(f"{pad}{self._h('History')}")
        print("")
        for job in content.job_history:
            print(f"{pad}  --------------------------------------")
            print(f"{pad}  Job Id: {job.job_id}")
            print(f"{pad}  Started: {job.start_ts_str}")
            print(f"{pad}  Ended: {job.end_ts_str}")
            print(f"{pad}  Duration: {job.duration_sec_str}")
            print(f"{pad}  Infra Sync Point: {job.sync_point}")
            if job.success is True:
                print(f"{pad}  Status: {self._gr('Success')}")
            else:
                print(f"{pad}  Status: {self._f('Fail')}")
            if job.error is not None:
                print(f"{pad}  Error: {self._gr(job.error)}")

            print("")

            if job.delta is None:
                print(f"{pad}{self._f('The job is missing a delta, never finished discovery.')}")
            else:
                if len(job.delta.added) > 0:
                    print(f"{pad}  {self._h('Added')}")
                    for added in job.delta.added:
                        vertex = infra.dag.get_vertex(added.uid)
                        if vertex is None:
                            print(f"{pad}  * Vertex {added.uid} does not exists.")
                        else:
                            if vertex.active is False:
                                print(f"{pad}  * Vertex {added.uid} is inactive.")
                            elif vertex.corrupt is True:
                                print(f"{pad}  * Vertex {added.uid} is corrupt.")
                            else:
                                content = DiscoveryObject.get_discovery_object(vertex)
                                print(f"{pad}  * {content.description}; Record UID: {content.record_uid}")
                    print("")

                if len(job.delta.changed) > 0:
                    print(f"{pad}  {self._h('Changed')}")
                    for changed in job.delta.changed:
                        vertex = infra.dag.get_vertex(changed.uid)
                        if vertex is None:
                            print(f"{pad}  * Vertex {changed.uid} does not exists.")
                        else:
                            if vertex.active is False:
                                print(f"{pad}  * Vertex {changed.uid} is inactive.")
                            elif vertex.corrupt is True:
                                print(f"{pad}  * Vertex {changed.uid} is corrupt.")
                            else:
                                content = DiscoveryObject.get_discovery_object(vertex)
                                print(f"{pad}  * {content.description}; Record UID: {content.record_uid}")
                                if changed.changes is not None:
                                    for k, v in changed.changes.items():
                                        print(f"{pad}    {k} = {v}")
                    print("")

                if len(job.delta.deleted) > 0:
                    print(f"{pad}  {self._h('Deleted')}")
                    for deleted in job.delta.deleted:
                        print(f"{pad}  * Removed vertex {deleted.uid}.")
                    print("")

    def _do_render_infra(self, params: KeeperParams, gateway_context: GatewayContext, filepath: str, graph_format: str,
                         debug_level: int = 0):

        infra = Infrastructure(record=gateway_context.configuration, params=params, logger=logging,
                               debug_level=debug_level)
        infra.load(sync_point=0)

        print("")
        dot_instance = infra.to_dot(
            graph_type=graph_format if graph_format != "raw" else "dot",
            show_only_active_vertices=False,
            show_only_active_edges=False
        )
        if graph_format == "raw":
            print(dot_instance)
        else:
            try:
                dot_instance.render(filepath)
                print(f"Infrastructure graph rendered to {self._gr(filepath)}")
            except Exception as err:
                print(self._f(f"Could not generate graph: {err}"))
                raise err
        print("")

    def _do_render_rl(self, params: KeeperParams, gateway_context: GatewayContext, filepath: str, graph_format: str,
                      debug_level: int = 0):

        rl = RecordLink(record=gateway_context.configuration, params=params, logger=logging, debug_level=debug_level)

        print("")
        dot_instance = rl.to_dot(
            graph_type=graph_format if graph_format != "raw" else "dot",
            show_only_active_vertices=False,
            show_only_active_edges=False
        )
        if graph_format == "raw":
            print(dot_instance)
        else:
            try:
                dot_instance.render(filepath)
                print(f"Record linking graph rendered to {self._gr(filepath)}")
            except Exception as err:
                print(self._f(f"Could not generate graph: {err}"))
                raise err
        print("")

    def _do_render_service(self, params: KeeperParams, gateway_context: GatewayContext, filepath: str,
                           graph_format: str, debug_level: int = 0):

        service = UserService(record=gateway_context.configuration, params=params, logger=logging,
                              debug_level=debug_level)

        print("")
        dot_instance = service.to_dot(
            graph_type=graph_format if graph_format != "raw" else "dot",
            show_only_active_vertices=False,
            show_only_active_edges=False
        )
        if graph_format == "raw":
            print(dot_instance)
        else:
            try:
                dot_instance.render(filepath)
                print(f"User service/tasks graph rendered to {self._gr(filepath)}")
            except Exception as err:
                print(self._f(f"Could not generate graph: {err}"))
                raise err
        print("")

    def _do_render_jobs(self, params: KeeperParams, gateway_context: GatewayContext, filepath: str,
                        graph_format: str, debug_level: int = 0):

        jobs = Jobs(record=gateway_context.configuration, params=params, logger=logging, debug_level=debug_level)

        print("")
        dot_instance = jobs.dag.to_dot()
        if graph_format == "raw":
            print(dot_instance)
        else:
            try:
                dot_instance.render(filepath)
                print(f"Job graph rendered to {self._gr(filepath)}")
            except Exception as err:
                print(self._f(f"Could not generate graph: {err}"))
                raise err
        print("")

    def _do_raw_text_list(self, params: KeeperParams, gateway_context: GatewayContext, graph_id: int = 0,
                          debug_level: int = 0):

        logging.debug(f"loading graph id {graph_id}, for record uid {gateway_context.configuration.record_uid}")

        conn = get_connection(params=params)
        dag = DAG(conn=conn, record=gateway_context.configuration, graph_id=graph_id, fail_on_corrupt=False,
                  logger=logging, debug_level=debug_level)
        dag.load(sync_point=0)
        print("")
        if dag.is_corrupt is True:
            print(f"{bcolors.FAIL}The graph is corrupt at Vertex UIDs: {', '.join(dag.corrupt_uids)}")
            print("")

        logging.debug("DAG DOT -------------------------------")
        logging.debug(str(dag.to_dot()))
        logging.debug("DAG DOT -------------------------------")

        line_start = {
            0: "",
            1: "* ",
            2: "- ",
            3: ". ",
        }

        color_func = {
            0: self._h,
            1: self._gr,
            2: self._bl,
            3: self._p
        }

        def _handle(current_vertex: DAGVertex, last_vertex: Optional[DAGVertex] = None, indent: int = 0):

            pad = ""
            if indent > 0:
                pad = "".ljust(4 * indent, ' ')

            ls = line_start.get(indent, "  ")
            cf = color_func.get(indent, self._p)
            text = f"{pad}{ls}{cf(current_vertex.uid)}"

            edge_types = []
            if last_vertex is not None:
                for edge in current_vertex.edges:  # type: DAGEdge
                    if edge.active is False:
                        continue
                    if edge.head_uid == last_vertex.uid:
                        edge_types.append(edge.edge_type.value)
            if len(edge_types) > 0:
                text += f"; edges: {', '.join(edge_types)}"

            if current_vertex.active is False:
                text += " " + self._f("Inactive")
            if current_vertex.corrupt is True:
                text += " " + self._f("Corrupt")

            print(text)

            if current_vertex.active is False:
                logging.debug(f"vertex {current_vertex.uid} is not active, will not get children.")
                return

            vertices = current_vertex.has_vertices()
            if len(vertices) == 0:
                logging.debug(f"vertex {current_vertex.uid} does not have any children.")
                return

            for vertex in vertices:
                _handle(vertex, current_vertex, indent=indent + 1)

        print("")
        _handle(dag.get_root)
        print("")

    def _do_raw_render_graph(self, params: KeeperParams, gateway_context: GatewayContext, filepath: str,
                             graph_format: str, graph_id: int = 0, debug_level: int = 0):

        conn = get_connection(params=params)
        dag = DAG(conn=conn, record=gateway_context.configuration, graph_id=graph_id, fail_on_corrupt=False,
                  logger=logging, debug_level=debug_level)
        dag.load(sync_point=0)
        dot = dag.to_dot(graph_format=graph_format)
        if graph_format == "raw":
            print(dot)
        else:
            try:
                dot.render(filepath)
                print(f"Graph rendered to {self._gr(filepath)}")
            except Exception as err:
                print(self._f(f"Could not generate graph: {err}"))
                raise err

        print("")

    def do_list(self, params: KeeperParams, gateway_context: GatewayContext, graph_type: str, debug_level: int = 0,
                indent: int = 0):
        list_func = getattr(self, f"_do_text_list_{graph_type}")
        list_func(params=params,
                  gateway_context=gateway_context,
                  debug_level=debug_level,
                  indent=indent)

    def execute(self, params: KeeperParams, **kwargs):

        gateway = kwargs.get("gateway")
        raw = kwargs.get("raw", False)
        graph_type = kwargs.get("graph_type")
        do_text_list = kwargs.get("do_text_list")
        do_render = kwargs.get("do_render")
        debug_level = int(kwargs.get("debug_level", 0))

        gateway_context = GatewayContext.from_gateway(params, gateway)
        if gateway_context is None:
            print(f"{bcolors.FAIL}Could not find the gateway configuration for {gateway}.")
            return

        if raw is True:
            if do_text_list is True:
                self._do_raw_text_list(params=params,
                                       gateway_context=gateway_context,
                                       graph_id=PAMDebugGraphCommand.graph_id_map.get(graph_type),
                                       debug_level=debug_level)
            if do_render is True:
                filepath = kwargs.get("filepath")
                graph_format = kwargs.get("format")
                self._do_raw_render_graph(params=params,
                                          gateway_context=gateway_context,
                                          filepath=filepath,
                                          graph_format=graph_format,
                                          graph_id=PAMDebugGraphCommand.graph_id_map.get(graph_type),
                                          debug_level=debug_level)
        else:
            if do_text_list is True:
                self.do_list(
                    params=params,
                    gateway_context=gateway_context,
                    graph_type=graph_type,
                    debug_level=debug_level
                )
            if do_render is True:
                filepath = kwargs.get("filepath")
                graph_format = kwargs.get("format")
                render_func = getattr(self, f"_do_render_{graph_type}")
                render_func(params=params,
                            gateway_context=gateway_context,
                            filepath=filepath,
                            graph_format=graph_format,
                            debug_level=debug_level)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import datetime
import hmac
import json
import logging
import os
import time
import urllib.parse
from itertools import product, groupby

from keeper_secrets_manager_core.utils import bytes_to_base64, url_safe_str_to_bytes

from typing import Sequence, List, Optional

from .base import Command, dump_report_data, user_choice, as_boolean
from . import record
from .. import api, utils, crypto, vault
from ..params import KeeperParams
from ..display import bcolors
from ..proto import APIRequest_pb2, record_pb2, enterprise_pb2
from ..error import KeeperApiError
from ..constants import get_abbrev_by_host
from ..utils import json_to_base64


available_ksm_commands = f"""
{bcolors.BOLD}Keeper Secrets Manager{bcolors.ENDC}
Commands to configure and manage the Keeper Secrets Manager platform.

  Usage:

  {bcolors.BOLD}View Applications:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager app list{bcolors.ENDC}

  {bcolors.BOLD}Get Application:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager app get {bcolors.OKBLUE}[APP NAME OR UID]{bcolors.ENDC}

  {bcolors.BOLD}Create Application:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager app create {bcolors.OKBLUE}[NAME]{bcolors.ENDC}

  {bcolors.BOLD}Remove Application:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager app remove {bcolors.OKBLUE}[APP NAME OR UID]{bcolors.ENDC}
    Options: 
      --purge : Remove the application and purge it from the trash
      --force : Do not prompt for confirmation

  {bcolors.BOLD}Grant User Access to Application (Share Application):{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager app share {bcolors.OKBLUE}[APP NAME OR UID]{bcolors.OKGREEN} --email {bcolors.OKBLUE}[USERNAME] {bcolors.ENDC}

  {bcolors.BOLD}Revoke User Access to Application (Unshare Application):{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager app unshare {bcolors.OKBLUE}[APP NAME OR UID]{bcolors.OKGREEN} --email {bcolors.OKBLUE}[USERNAME]{bcolors.ENDC}

  {bcolors.BOLD}Add Client Device:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager client add --app {bcolors.OKBLUE}[APP NAME OR UID] {bcolors.OKGREEN}--unlock-ip{bcolors.ENDC}
    Options: 
      --name [CLIENT NAME] : Name of the client (Default: Random 10 characters string)
      --first-access-expires-in-min [MIN] : First time access expiration (Default 60, Max 1440)
      --access-expire-in-min [MIN] : Client access expiration (Default: no expiration)
      --unlock-ip : Does not lock IP address to first requesting device
      --count [NUM] : Number of tokens to generate (Default: 1)
      --config-init [json, b64 or k8s] : Initialize configuration string from a one-time token

  {bcolors.BOLD}Remove Client Device:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager client remove --app {bcolors.OKBLUE}[APP NAME OR UID] {bcolors.OKGREEN}--client {bcolors.OKBLUE}[NAME OR ID]{bcolors.ENDC}
    Options: 
      --force : Do not prompt for confirmation
      --client : Client name or ID. Provide `*` or `all` to delete all clients at once
      
  {bcolors.BOLD}Add Secret to Application:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager share add --app {bcolors.OKBLUE}[APP NAME OR UID] {bcolors.OKGREEN}--secret {bcolors.OKBLUE}[RECORD OR SHARED FOLDER UID]{bcolors.ENDC}
    Options: 
      --editable : Allow secrets to be editable by the client

  {bcolors.BOLD}Remove Secret from Application:{bcolors.ENDC}
  {bcolors.OKGREEN}secrets-manager share remove --app {bcolors.OKBLUE}[APP NAME OR UID] {bcolors.OKGREEN}--secret {bcolors.OKBLUE}[RECORD OR SHARED FOLDER UID]{bcolors.ENDC}

  -----
  Note: If the UID you are using contains a dash (-) in the beginning, the value should be wrapped 
  in quotes and prepended with an equal sign. For example:
  {bcolors.OKGREEN}secrets-manager share add --app={bcolors.OKBLUE}"-fwZjKGbKnZCo1Fh8gsf5w"{bcolors.OKGREEN} --secret={bcolors.OKBLUE}"-FcesCt6YXcJzpHWWRgoDA"{bcolors.ENDC}

  To learn about Keeper Secrets Manager visit:
  {bcolors.WARNING}https://docs.keeper.io/secrets-manager/{bcolors.ENDC}

"""

ksm_parser = argparse.ArgumentParser(prog='secrets-manager', description='Keeper Secrets Management (KSM) Commands',
                                     add_help=False)
ksm_parser.add_argument('command', type=str, action='store', nargs="*",
                    help='One of: "app list", "app get", "app create", "app remove", "app share", "app unshare", ' +
                             '"client add", "client remove", "share add" or "share remove"')
ksm_parser.add_argument('--secret', '-s', type=str, action='append', required=False,
                        help='Record UID')
ksm_parser.add_argument('--app', '-a', type=str, action='store', required=False,
                        help='Application Name or UID')
ksm_parser.add_argument('--client', '-i', type=str, dest='client_names_or_ids', action='append', required=False,
                        help='Client Name or ID')
ksm_parser.add_argument('--first-access-expires-in-min', '-x', type=int, dest='firstAccessExpiresIn', action='store',
                        help='Time for the first request to expire in minutes from the time when this command is '
                             'executed. Maximum 1440 minutes (24 hrs). Default: 60',
                        default=60)
ksm_parser.add_argument('--access-expire-in-min', '-p', type=int, dest='accessExpireInMin', action='store',
                        help='Time interval that this client can access the KSM application. After this time, access '
                             'is denied. Time is entered in minutes starting from the time when command is executed. '
                             'Default: Not expiration')

ksm_parser.add_argument('--count', '-c', type=int, dest='count', action='store',
                        help='Number of tokens to return. Default: 1', default=1)
ksm_parser.add_argument('--help', '-h', dest='helpflag', action="store_true", help='Display help')
ksm_parser.add_argument('--editable', '-e', action='store_true', required=False,
                        help='Is this share going to be editable or not.')
ksm_parser.add_argument('--unlock-ip', '-l', dest='unlockIp', action='store_true', help='Unlock IP Address.')
ksm_parser.add_argument('--return-tokens', dest='returnTokens', action='store_true', help='Return Tokens')
ksm_parser.add_argument('--name', '-n', type=str, dest='name', action='store', help='client name')
ksm_parser.add_argument('--purge', dest='purge', action='store_true',
                        help='remove the record from all folders and purge it from the trash')
ksm_parser.add_argument('-f', '--force', dest='force', action='store_true', help='do not prompt')
ksm_parser.add_argument('--config-init', type=str, dest='config_init', action='store',
                        help='Initialize client config')    # json, b64, file
# Application sharing options
ksm_parser.add_argument('--email', action='store', type=str, dest='email', help='Email of user to grant / remove application access to / from')
# Disable sharing apps w/ admin permissions for now
# ksm_parser.add_argument('--admin', action='store_true', help='Allow share recipient to manage application')
ksm_parser.add_argument('--format', dest='format', action='store', choices=['table', 'json'], default='table',
                        help='Output format (table, json)')



def ms_to_str(ms, frmt='%Y-%m-%d %H:%M:%S'):
    dt = datetime.datetime.fromtimestamp(ms // 1000)
    df_frmt_str = dt.strftime(frmt)

    return df_frmt_str


class KSMCommand(Command):

    CLIENT_SHORT_ID_LENGTH = 8

    def get_parser(self):
        return ksm_parser

    def execute(self, params, **kwargs):

        ksm_command = kwargs.get('command') # type: List[str]
        ksm_helpflag = kwargs.get('helpflag')

        if len(ksm_command) == 0 or ksm_helpflag:
            print(available_ksm_commands)
            return

        ksm_obj = ksm_command[0]
        ksm_action = ksm_command[1] if len(ksm_command) > 1 else None

        if ksm_obj in ['help', 'h']:
            print(available_ksm_commands)
            return

        if ksm_obj == 'apps' or \
                (ksm_obj in ['app', 'apps'] and ksm_action == 'list'):
            format_type = kwargs.get('format', 'table')
            return KSMCommand.print_all_apps_records(params, format_type)

        if ksm_obj == 'clients' or (ksm_obj in ['client', 'clients'] and ksm_action == 'list'):
            print(bcolors.WARNING + "Listing clients is not available" + bcolors.ENDC)
            return

        if ksm_obj in ['app', 'apps'] and ksm_action == 'get':

            if len(ksm_command) != 3:
                print(f"{bcolors.WARNING}Application name is required.\n  " +
                      f"Example: {bcolors.OKGREEN}secrets-manager app get {bcolors.OKBLUE}MyApp{bcolors.ENDC}")
                return

            ksm_app_uid_or_name = ksm_command[2]

            ksm_app = KSMCommand.get_app_record(params, ksm_app_uid_or_name)

            if not ksm_app:
                format_type = kwargs.get('format', 'table')
                if format_type == 'json':
                    return json.dumps({"error": f"Application '{ksm_app_uid_or_name}' not found."})
                else:
                    print((bcolors.WARNING + "Application '%s' not found." + bcolors.ENDC) % ksm_app_uid_or_name)
                return

            format_type = kwargs.get('format', 'table')
            result = KSMCommand.get_and_print_app_info(params, ksm_app.get('record_uid'), format_type)
            if format_type == 'json' and result:
                return result
            return

        if ksm_obj in ['client'] and ksm_action == 'get':
            print(bcolors.WARNING + "Viewing clients is not available" + bcolors.ENDC)
            return

        if ksm_obj in ['app', 'apps'] and ksm_action in ['add', 'create']:
            if len(ksm_command) < 3:
                print(
                    f'''{bcolors.WARNING}Application name is missing.{bcolors.ENDC}\n'''
                    f'''\tEx: {bcolors.OKGREEN}secrets-manager app create {bcolors.OKBLUE}MyApp{bcolors.ENDC}'''
                )
                return

            ksm_app_name = ksm_command[2]

            force_to_add = False    # TODO: externalize this
            format_type = kwargs.get('format', 'table')

            result = KSMCommand.add_new_v5_app(params, ksm_app_name, force_to_add, format_type)
            if format_type == 'json' and result:
                return result
            return

        if ksm_obj in ['app', 'apps'] and ksm_action in ['remove', 'rem', 'rm']:
            app_name_or_uid = ksm_command[2]
            purge = kwargs.get('purge')
            force = kwargs.get('force')

            KSMCommand.remove_v5_app(params=params, app_name_or_uid=app_name_or_uid, purge=purge, force=force)

            return

        if ksm_obj in ['app', 'apps'] and ksm_action in ['share', 'unshare']:
            app_name_or_uid = kwargs.get('app') or ksm_command[2] if len(ksm_command) == 3 else None
            if not app_name_or_uid:
                print(
                    f'''{bcolors.WARNING}Application name is missing.{bcolors.ENDC}\n'''
                    f'''\tEx: {bcolors.OKGREEN}secrets-manager app {ksm_action} {bcolors.OKBLUE}--app=MyApp{bcolors.OKGREEN} --email=myemail@mydomain.com{bcolors.ENDC}'''
                )
                return
            email = kwargs.get('email')
            unshare = 'un' in ksm_action
            is_admin = kwargs.get('admin', False)
            if not email:
                print(
                    f'''{bcolors.WARNING}Email is missing.{bcolors.ENDC}\n'''
                    f'''\tEx: {bcolors.OKGREEN}secrets-manager app {ksm_action} --app=MyApp {bcolors.OKBLUE}--email=myemail@mydomain.com{bcolors.ENDC}'''
                )
                return

            KSMCommand.share_app(params, app_name_or_uid, email, is_admin=is_admin, unshare=unshare)
            return

        if ksm_obj in ['share', 'secret'] and ksm_action is None:
            print("  Add Secret to the App\n\n"
                  + bcolors.OKGREEN + "    secrets-manager share add --app " + bcolors.OKBLUE + "[APP NAME or APP UID]"
                  + bcolors.OKGREEN + " --secret " + bcolors.OKBLUE + "[SECRET UID or SHARED FOLDER UID]"
                  + bcolors.OKGREEN + " --editable" + bcolors.ENDC + "\n")
            return

        if ksm_obj in ['share', 'secret'] and ksm_action in ['add', 'create']:

            app_name_or_uid = kwargs.get('app')
            secret_uid = kwargs.get('secret')   # TODO: Allow multiple secrets
            is_editable = kwargs.get('editable')

            if not secret_uid:
                print(bcolors.WARNING + "\nRecord or Shared Folder UID is required." + bcolors.ENDC)
                print(f"Example to add secret:"
                      + bcolors.OKGREEN + " secrets-manager share add --app " + bcolors.OKBLUE + "[APP NAME or APP UID]"
                      + bcolors.OKGREEN + " --secret " + bcolors.OKBLUE + "[SECRET UID or SHARED FOLDER UID]"
                      + bcolors.OKGREEN + " --editable" + bcolors.ENDC + "\n")
                return

            KSMCommand.add_app_share(params, secret_uid, app_name_or_uid, is_editable)
            return

        if ksm_obj in ['share', 'secret'] and ksm_action in ['remove', 'rem', 'rm']:
            app_name_or_uid = kwargs['app'] if 'app' in kwargs else None
            secret_uids = kwargs.get('secret')

            KSMCommand.remove_share(params, app_name_or_uid, secret_uids)
            return

        if ksm_obj in ['client', 'c']:
            app_name_or_uid = kwargs['app'] if 'app' in kwargs else None

            if not app_name_or_uid:
                print(bcolors.WARNING + "App name is required" + bcolors.ENDC)
                print(f"{bcolors.OKGREEN}  secrets-manager client add "
                      f"{bcolors.OKGREEN}--app {bcolors.OKBLUE}[APP NAME or APP UID] "
                      f"{bcolors.OKGREEN}--secret {bcolors.OKBLUE}[SECRET UID or SHARED FOLDER UID] "
                      f"{bcolors.OKGREEN}--name {bcolors.OKBLUE}[CLIENT NAME] "
                      f"{bcolors.OKGREEN}--config-init [{bcolors.OKBLUE}json{bcolors.OKGREEN}, "
                      f"{bcolors.OKBLUE}b64{bcolors.OKGREEN} or "
                      f"{bcolors.OKBLUE}k8s{bcolors.OKGREEN}]")
                return

            if ksm_action in ['add', 'create']:
                count = kwargs.get('count')
                unlock_ip = kwargs.get('unlockIp')

                client_name = kwargs.get('name')
                config_init = kwargs.get('config_init')

                first_access_expire_on = kwargs.get('firstAccessExpiresIn')
                access_expire_in_min = kwargs.get('accessExpireInMin')

                is_return_tokens = kwargs.get('returnTokens')

                tokens_and_device = KSMCommand.add_client(
                    params, app_name_or_uid, count, unlock_ip, first_access_expire_on, access_expire_in_min,
                    client_name=client_name, config_init=config_init, client_type=enterprise_pb2.GENERAL)

                if config_init:
                    tokens_only = [d['config'] for d in tokens_and_device]
                else:
                    tokens_only = [d['oneTimeToken'] for d in tokens_and_device]

                return ', '.join(tokens_only) if is_return_tokens else None

            if ksm_action in ['remove', 'rem', 'rm']:
                client_names_or_ids = kwargs.get('client_names_or_ids')
                if not client_names_or_ids:
                    return

                force = kwargs.get('force')

                if len(client_names_or_ids) == 1 and client_names_or_ids[0] in ['*', 'all']:
                    KSMCommand.remove_all_clients(params, app_name_or_uid, force)
                else:
                    KSMCommand.remove_client(params, app_name_or_uid, client_names_or_ids)

                return

        print(f"{bcolors.WARNING}Unknown combination of KSM commands. " +
              f"Type 'secrets-manager' for more details'{bcolors.ENDC}")

    @staticmethod
    def share_app(params, app_name_or_uid, email, is_admin=False, unshare=False):
        # type: (KeeperParams, str, List[str], Optional[bool], Optional[bool]) -> None
        app_rec = KSMCommand.get_app_record(params, app_name_or_uid)
        if app_rec is None:
            logging.warning('Application "%s" not found.' % app_name_or_uid)
            return

        # For now, disable sharing app w/ edit + share permissions
        is_admin = False

        app_uid = app_rec.get('record_uid', '')
        sr_action = 'revoke' if unshare else 'grant'
        rec_perms = dict(can_edit=is_admin and not unshare, can_share=is_admin and not unshare)
        users = [email]
        share_rec_args = dict(**rec_perms, action=sr_action, email=users)

        # (Un)Share application record
        from .register import ShareRecordCommand
        share_rec_cmd = ShareRecordCommand()
        share_rec_cmd.execute(params, record=app_uid, **share_rec_args)

        api.sync_down(params)
        KSMCommand.update_secrets_user_permissions(params, app_uid, removed=unshare and email or None)

    @staticmethod
    def update_secrets_user_permissions(params, app_uid, removed=None):   # type: (KeeperParams, str, Optional[str] ) -> None
        # Get app user-permissions
        api.get_record_shares(params, [app_uid])
        app_rec = KSMCommand.get_app_record(params, app_uid)
        if app_rec is None:
            logging.warning('Application "%s" not found.' % app_uid)
            return

        user_perms = app_rec.get('shares', {}).get('user_permissions', [])
        sf_perm_keys = ('manage_users', 'manage_records')
        rec_perm_keys = ('can_edit', 'can_share')

        # Grant app-users access to shares as needed
        app_info = KSMCommand.get_app_info(params, app_uid)
        share_uids = [utils.base64_url_encode(s.secretUid) for ai in app_info for s in (ai.shares or [])]
        shared_recs = [uid for uid in share_uids if uid in params.record_cache]
        shared_folders = [uid for uid in share_uids if uid in params.shared_folder_cache]

        # Exclude un-shareable secrets
        api.get_record_shares(params, shared_recs)
        get_sf_permissions = lambda uid: params.shared_folder_cache.get(uid, {}).get('users', [])
        get_rec_permissions = lambda uid: params.record_cache.get(uid, {}).get('shares', {}).get('user_permissions', {})
        is_sf_admin = lambda uid: params.user in api.get_share_admins_for_shared_folder(params, uid)
        is_rec = lambda uid: uid in shared_recs
        get_perms = lambda uid: get_rec_permissions(uid) if is_rec(uid) else get_sf_permissions(uid)
        get_share_user_perms = lambda user, uid: next((x for x in get_perms(uid) if x.get('username') == user), {})
        get_perm = lambda uid, name: get_share_user_perms(params.user, uid).get(name, False)
        is_shareable = lambda uid: (get_perm(uid, 'shareable') or get_perm(uid, 'share_admin')
                                    or  get_perm(uid, 'manage_users') or is_sf_admin(uid))
        shared_recs = [uid for uid in shared_recs if is_shareable(uid)]
        shared_folders = [uid for uid in shared_folders if is_shareable(uid)]

        def share_needs_update(user, share_uid, elevated):
            if is_removed:
                # Allow user to retain access to app secrets after removal from app
                return False
            else:
                # Share secret w/ user only if they lack access
                return not get_share_user_perms(user, share_uid)

        admins = [up.get('username') for up in user_perms if up.get('editable')]
        admins = [x for x in admins if x != params.user]    # Exclude current user
        viewers = [up.get('username') for up in user_perms if not up.get('editable')]
        removed = [removed] if removed is not None else []
        app_users_map = dict(admins=admins,viewers=viewers, removed=removed)

        from .register import ShareRecordCommand, ShareFolderCommand

        get_sf = lambda uid: params.shared_folder_cache.get(uid, {})
        user_needs_update = lambda u, adm: any(share_needs_update(u, uid, adm) for uid in share_uids)

        def group_by_app_share(products):
            first_element = lambda x: x[0]
            products = sorted(products, key=first_element)
            products = groupby(products, key=first_element)
            return {uid: [user for _, user in pair] for uid, pair in products}

        sf_requests = []
        rec_requests = []
        for group, users in app_users_map.items():
            is_admin, is_removed = group == 'admins', group == 'removed'
            users = [u for u in users if user_needs_update(u, is_admin)]
            sf_action = 'remove' if is_removed else 'grant'
            rec_action = 'revoke' if is_removed else 'grant'

            # Grant app-users access (using folder's default permissions) to folder shares if needed
            get_sf_perms = lambda sf: {perm: 'on' if sf.get(f'default_{perm}', False) else 'off' for perm in sf_perm_keys}
            get_sf_args = lambda u, uid: dict(action=sf_action, user=u, **get_sf_perms(get_sf(uid)))
            prep_sf_rq = lambda u, uid: ShareFolderCommand.prepare_request(params, get_sf_args(u, uid), get_sf(uid), u, [], [])
            sf_updates = {(sf, user) for sf, user in product(shared_folders, users) if share_needs_update(user, sf, is_admin)}
            sf_updates = group_by_app_share(sf_updates)
            sf_requests.append([prep_sf_rq(users, uid) for uid, users in sf_updates.items() if users])

            # Grant app-users read-only access to record shares if not already shared
            get_rec_perms = lambda uid: {perm: False for perm in rec_perm_keys}
            get_rec_args = lambda user, uid: dict(action=rec_action, email=user, record=uid, **get_rec_perms(uid))
            prep_rec_rq = lambda user, uid: ShareRecordCommand.prep_request(params, get_rec_args(user, uid))
            rec_updates = {(rec, user) for rec, user in product(shared_recs, users) if share_needs_update(user, rec, is_admin)}
            rec_updates = group_by_app_share(rec_updates)
            rec_requests.extend([prep_rec_rq(users, rec) for rec, users in rec_updates.items() if users])
            rec_requests = [rq for rq in rec_requests if rq]

        ShareFolderCommand.send_requests(params, sf_requests)
        ShareRecordCommand.send_requests(params, rec_requests)


    @staticmethod
    def add_app_share(params, secret_uids, app_name_or_uid, is_editable):

        rec_cache_val = KSMCommand.get_app_record(params, app_name_or_uid)
        if rec_cache_val is None:
            logging.warning('Application "%s" not found.' % app_name_or_uid)
            return

        app_record_uid = rec_cache_val.get('record_uid')
        master_key = rec_cache_val.get('record_key_unencrypted')

        KSMCommand.share_secret(
            params=params,
            app_uid=app_record_uid,
            master_key=master_key,
            secret_uids=secret_uids,
            is_editable=is_editable
        )

        # Update user-permissions for new app share
        api.sync_down(params)
        KSMCommand.update_secrets_user_permissions(params, app_record_uid)

    @staticmethod
    def record_data_as_dict(record_dict):
        data_json_str = record_dict.get('data_unencrypted').decode("utf-8")
        data_dict = json.loads(data_json_str)
        return data_dict

    @staticmethod
    def print_all_apps_records(params, format_type='table'):

        if format_type == 'table':
            print(f"\n{bcolors.BOLD}List all Secrets Manager Applications{bcolors.ENDC}\n")
        
        rs = api.communicate_rest(params, None, 'vault/get_applications_summary',
                                  rs_type=APIRequest_pb2.GetApplicationsSummaryResponse)

        app_summary = {utils.base64_url_encode(x.appRecordUid): {
            'last_access': x.lastAccess,
            'record_shares': x.recordShares,
            'folder_shares': x.folderShares,
            'folder_records': x.folderRecords,
            'client_count': x.clientCount,
        } for x in rs.applicationSummary}

        if format_type == 'json':
            apps_table_fields = ['app_name', 'app_uid', 'records', 'folders', 'devices', 'last_access']
        else:
            apps_table_fields = [f'{bcolors.OKGREEN}App Name{bcolors.ENDC}', f'{bcolors.OKBLUE}App UID{bcolors.ENDC}',
                                 'Records', 'Folders', 'Devices', 'Last Access']
        
        apps_table = []
        for app_uid in app_summary:
            app = app_summary[app_uid]
            app_record = vault.KeeperRecord.load(params, app_uid)
            if isinstance(app_record, vault.ApplicationRecord):
                la = app['last_access']
                if la > 0:
                    last_access = datetime.datetime.fromtimestamp(la // 1000)
                    last_access_str = last_access.strftime('%Y-%m-%d %H:%M:%S') if format_type == 'json' else last_access
                else:
                    last_access_str = None
                
                if format_type == 'json':
                    row = [app_record.title, app_uid, app['folder_records'], app['folder_shares'], 
                           app['client_count'], last_access_str]
                else:
                    row = [f'{bcolors.OKGREEN}{app_record.title}{bcolors.ENDC}', f'{bcolors.OKBLUE}{app_uid}{bcolors.ENDC}',
                           app['folder_records'], app['folder_shares'], app['client_count'], last_access_str]
                apps_table.append(row)

        apps_table.sort(key=lambda x: x[0].lower() if format_type == 'json' else x[0].replace('\x1b[92m', '').replace('\x1b[0m', '').lower())

        if len(apps_table) == 0:
            if format_type == 'json':
                print(json.dumps({"applications": [], "message": "No Applications to list."}))
            else:
                print(f'{bcolors.WARNING}No Applications to list.{bcolors.ENDC}\n\n'
                      f'To create new application, use command {bcolors.OKGREEN}secrets-manager app '
                      f'create {bcolors.OKBLUE}[NAME]{bcolors.ENDC}')
        else:
            return dump_report_data(apps_table, apps_table_fields, fmt=format_type)

        if format_type == 'table':
            print("")

    @staticmethod
    def get_app_info(params, app_uid):   # type: (KeeperParams, str) -> Sequence[APIRequest_pb2.AppInfo]
        rq = APIRequest_pb2.GetAppInfoRequest()
        rq.appRecordUid.append(utils.base64_url_decode(app_uid))
        rs = api.communicate_rest(params, rq, 'vault/get_app_info', rs_type=APIRequest_pb2.GetAppInfoResponse)
        return rs.appInfo

    @staticmethod
    def get_sm_app_record_by_uid(params, uid):
        rec = params.record_cache.get(uid)

        if rec.get('version') != 5:
            raise Exception(f'Record {uid} is not a Secrets Manager application')

        data_json_str = rec.get('data_unencrypted').decode("utf-8")
        data_dict = json.loads(data_json_str)

        return data_dict

    @staticmethod
    def get_and_print_app_info(params, uid, format_type='table'):

        app_info = KSMCommand.get_app_info(params, uid)

        def shorten_client_id(all_clients, original_id, number_of_characters):

            new_id = original_id[0:number_of_characters]

            res = list(filter(lambda x: utils.base64_url_encode(x.clientId).startswith(new_id), all_clients))
            if len(res) == 1 or new_id == original_id:
                return new_id
            else:
                return shorten_client_id(all_clients, original_id, number_of_characters+1)

        if len(app_info) == 0:
            if format_type == 'json':
                return json.dumps({"error": "No Secrets Manager Applications returned."})
            else:
                print(bcolors.WARNING + 'No Secrets Manager Applications returned.' + bcolors.ENDC)
            return
        else:
            result_data = []
            for ai in app_info:

                app_uid_str = utils.base64_url_encode(ai.appRecordUid)

                app = KSMCommand.get_sm_app_record_by_uid(params, app_uid_str)
                
                app_data = {
                    "app_name": app.get("title"),
                    "app_uid": app_uid_str,
                    "client_devices": [],
                    "shares": []
                }
                
                if format_type == 'table':
                    print(f'\nSecrets Manager Application\n'
                        f'App Name: {app.get("title")}\n'
                        f'App UID: {app_uid_str}')

                client_devices = [x for x in ai.clients if x.appClientType == enterprise_pb2.GENERAL]
                if len(client_devices) > 0:
                    client_count = 1
                    for c in client_devices:
                        client_id = utils.base64_url_encode(c.clientId)
                        current_milli_time = round(time.time() * 1000)
                        
                        created_on_ts = ms_to_str(c.createdOn)
                        first_access_ts = None if c.firstAccess == 0 else ms_to_str(c.firstAccess)
                        last_access_ts = None if c.lastAccess == 0 else ms_to_str(c.lastAccess)
                        
                        if c.accessExpireOn == 0:
                            expire_access_ts = None
                            expire_status = "never"
                        elif c.accessExpireOn <= current_milli_time:
                            expire_access_ts = ms_to_str(c.accessExpireOn)
                            expire_status = "expired"
                        else:
                            expire_access_ts = ms_to_str(c.accessExpireOn)
                            expire_status = "active"
                        
                        short_client_id = shorten_client_id(ai.clients, client_id, KSMCommand.CLIENT_SHORT_ID_LENGTH)
                        
                        client_device_data = {
                            "device_name": c.id,
                            "short_id": short_client_id,
                            "client_id": client_id,
                            "created_on": created_on_ts,
                            "expires_on": expire_access_ts,
                            "expire_status": expire_status,
                            "first_access": first_access_ts,
                            "last_access": last_access_ts,
                            "ip_lock_enabled": c.lockIp,
                            "ip_address": c.ipAddress if c.ipAddress else None
                        }
                        app_data["client_devices"].append(client_device_data)
                        
                        if format_type == 'table':
                            created_on = f'{bcolors.OKGREEN}{created_on_ts}{bcolors.ENDC}'
                            first_access = f'{bcolors.WARNING}Never{bcolors.ENDC}' if c.firstAccess == 0 else f'{bcolors.OKGREEN}{first_access_ts}{bcolors.ENDC}'
                            last_access = f'{bcolors.WARNING}Never{bcolors.ENDC}' if c.lastAccess == 0 else f'{bcolors.OKGREEN}{last_access_ts}{bcolors.ENDC}'
                            lock_ip = f'{bcolors.OKGREEN}Enabled{bcolors.ENDC}' if c.lockIp else f'{bcolors.WARNING}Disabled{bcolors.ENDC}'
                            
                            if expire_status == "never":
                                expire_access = f'{bcolors.OKGREEN}Never{bcolors.ENDC}'
                            elif expire_status == "expired":
                                expire_access = f'{bcolors.FAIL}{expire_access_ts}{bcolors.ENDC}'
                            else:
                                expire_access = f'{bcolors.WARNING}{expire_access_ts}{bcolors.ENDC}'

                            client_devices_str = f"\n{bcolors.BOLD}Client Device {client_count}{bcolors.ENDC}\n" \
                                                f"=============================\n" \
                                                f'  Device Name: {bcolors.OKGREEN}{c.id}{bcolors.ENDC}\n' \
                                                f'  Short ID: {bcolors.OKGREEN}{short_client_id}{bcolors.ENDC}\n' \
                                                f'  Created On: {created_on}\n' \
                                                f'  Expires On: {expire_access}\n' \
                                                f'  First Access: {first_access}\n' \
                                                f'  Last Access: {last_access}\n' \
                                                f'  IP Lock: {lock_ip}\n' \
                                                f'  IP Address: {client_device_data["ip_address"] or "--"}'

                            print(client_devices_str)
                        client_count += 1

                else:
                    if format_type == 'table':
                        print(f'\n\t{bcolors.WARNING}No client devices registered for this Application{bcolors.ENDC}')

                if format_type == 'table':
                    print(bcolors.BOLD + "\nApplication Access\n" + bcolors.ENDC)

                if ai.shares:
                    recs = params.record_cache

                    for s in ai.shares:
                        uid_str = utils.base64_url_encode(s.secretUid)
                        sht = APIRequest_pb2.ApplicationShareType.Name(s.shareType)
                        
                        share_data = {
                            "share_type": sht,
                            "uid": uid_str,
                            "editable": s.editable
                        }

                        if sht == 'SHARE_TYPE_RECORD':
                            rec = recs.get(uid_str)
                            if rec:
                                record_data_dict = KSMCommand.record_data_as_dict(rec)
                                share_data["title"] = record_data_dict.get('title')
                                share_data["type"] = "RECORD"
                        elif sht == 'SHARE_TYPE_FOLDER':
                            if uid_str in params.shared_folder_cache:
                                cached_sf = params.shared_folder_cache[uid_str]
                                share_data["title"] = cached_sf.get('name_unencrypted')
                                share_data["type"] = "FOLDER"
                            else:
                                continue
                        else:
                            logging.warning("Unknown Share Type %s" % sht)
                            continue

                        app_data["shares"].append(share_data)
                    
                    if format_type == 'table' and ai.shares:
                        shares_table_fields = ['Share Type', 'UID', 'Title', 'Permissions']
                        shares_table = []
                        
                        for share in app_data["shares"]:
                            uid_str_c = bcolors.OKBLUE + share["uid"] + bcolors.ENDC
                            editable_status_color = bcolors.OKGREEN if share["editable"] else bcolors.WARNING
                            editable_status = editable_status_color + ("Editable" if share["editable"] else "Read-Only") + bcolors.ENDC
                            
                            row = [
                                share["type"],
                                uid_str_c,
                                share.get("title", ""),
                                editable_status]
                            shares_table.append(row)

                        shares_table.sort(key=lambda x: x[2].lower())
                        dump_report_data(shares_table, shares_table_fields, fmt='table')
                        print()
                else:
                    if format_type == 'table':
                        print('\tThere are no shared secrets to this application')
                
                result_data.append(app_data)
            
        if format_type == 'json':
            if len(result_data) == 1:
                return json.dumps(result_data[0], indent=2)
            else:
                return json.dumps({"applications": result_data}, indent=2)

    @staticmethod
    def share_secret(params, app_uid, master_key, secret_uids, is_editable=False):

        app_shares = []

        added_secret_uids_type_pairs = []

        for uid in secret_uids:
            is_record = uid in params.record_cache
            is_shared_folder = api.is_shared_folder(params, uid)

            if is_record:
                rec = params.record_cache[uid]
                share_key_decrypted = rec['record_key_unencrypted']
                share_type = 'SHARE_TYPE_RECORD'
            elif is_shared_folder:
                cached_sf = params.shared_folder_cache[uid]
                shared_folder_key_unencrypted = cached_sf.get('shared_folder_key_unencrypted')
                share_key_decrypted = shared_folder_key_unencrypted
                share_type = 'SHARE_TYPE_FOLDER'
            else:
                print(f"""{bcolors.WARNING}\tUID="{uid}" is not a Record nor Shared Folder. Only individual records or 
                Shared Folders can be added to the application.{bcolors.ENDC} Make sure your local cache is up to date by
                running 'sync-down' command and trying again.""")

                continue

            added_secret_uids_type_pairs.append((uid, share_type))

            encrypted_secret_key = crypto.encrypt_aes_v2(share_key_decrypted, master_key)

            app_share = APIRequest_pb2.AppShareAdd()
            app_share.secretUid = utils.base64_url_decode(uid)
            app_share.shareType = APIRequest_pb2.ApplicationShareType.Value(share_type)
            app_share.encryptedSecretKey = encrypted_secret_key
            app_share.editable = is_editable

            app_shares.append(app_share)

        if len(added_secret_uids_type_pairs) == 0:
            return

        app_share_add_rq = APIRequest_pb2.AddAppSharesRequest()
        app_share_add_rq.appRecordUid = utils.base64_url_decode(app_uid)
        app_share_add_rq.shares.extend(app_shares)

        try:
            api.communicate_rest(params, app_share_add_rq, 'vault/app_share_add')
            print(bcolors.OKGREEN + f'\nSuccessfully added secrets to app uid={app_uid}, '
                                    f'editable=' + bcolors.BOLD + f'{is_editable}:' + bcolors.ENDC)
            print('\n'.join(map(lambda x: ('\t' + str(x[0])) + ' ' + ('Record' if ('RECORD' in str(x[1])) else 'Shared Folder'), added_secret_uids_type_pairs)))
            print('\n')
            return True
        except KeeperApiError as kae:
            if kae.message == 'Duplicate share, already added':
                logging.error("One of the secret UIDs is already shared to this application. "
                              "Please remove already shared UIDs from your command and try again.")
            else:
                raise kae
        return False

    @staticmethod
    def get_app_record(params, app_name_or_uid):

        for rec_cache_val in params.record_cache.values():

            if rec_cache_val.get('version') == 5:
                r_uid = rec_cache_val.get('record_uid')
                r_unencr_json_data = rec_cache_val.get('data_unencrypted').decode('utf-8')
                r_unencr_dict = json.loads(r_unencr_json_data)

                if r_unencr_dict.get('title') == app_name_or_uid or r_uid == app_name_or_uid:
                    return rec_cache_val

        return None

    @staticmethod
    def search_app_records(params, search_str):

        search_results_rec_data = []

        for record_uid in params.record_cache:
            rec = params.record_cache[record_uid]

            if rec.get('version') == 5:
                data = json.loads(rec.get('data_unencrypted'))
                rec_uid = rec.get('record_uid')
                rec_title = data.get('title')

                if rec_uid == search_str.strip() or rec_title.lower() == search_str.strip().lower():
                    search_results_rec_data.append(
                        {
                            'uid': rec_uid,
                            'data': data
                        })

        return search_results_rec_data

    @staticmethod
    def remove_v5_app(params, app_name_or_uid, purge, force):

        app = KSMCommand.get_app_record(params, app_name_or_uid)

        if not app:
            logging.warning('Application "%s" not found.' % app_name_or_uid)
            return
        app_uid = app.get('record_uid')

        app_info = KSMCommand.get_app_info(params, app_uid)

        clients_count = len(app_info[0].clients)
        shared_folders_count = sum(map(lambda s: s.shareType == 1, app_info[0].shares))
        shared_records_count = sum(map(lambda s: s.shareType == 0, app_info[0].shares))

        if not force:
            print("This Application (uid: %s) has %d client(s), %d shared folder(s), and %d record(s)."
                  % (app_uid, clients_count, shared_folders_count, shared_records_count))
            uc = user_choice('\tAre you sure you want to delete this application?', 'yn', default='n')
            if uc.lower() != 'y':
                return

        logging.info("Removed Application uid: %s" % app_uid)

        cmd = record.RecordRemoveCommand()
        cmd.execute(params, purge=purge, force=True, record=app_uid)

    @staticmethod
    def add_new_v5_app(params, app_name, force_to_add=False, format_type='table'):

        logging.debug("Creating new KSM Application named '%s'" % app_name)

        found_app = KSMCommand.get_app_record(params, app_name)
        if (found_app is not None) and (found_app is not force_to_add):
            if format_type == 'json':
                return json.dumps({"error": f'Application with the same name "{app_name}" already exists.'})
            else:
                logging.warning('Application with the same name "%s" already exists.' % app_name)
            return

        app_record_data = {
            'title': app_name,
            'type': 'app'
        }

        data_json = json.dumps(app_record_data)
        record_key_unencrypted = utils.generate_aes_key()
        record_key_encrypted = crypto.encrypt_aes_v2(record_key_unencrypted, params.data_key)

        app_record_uid_str = api.generate_record_uid()
        app_record_uid = utils.base64_url_decode(app_record_uid_str)

        data = data_json.decode('utf-8') if isinstance(data_json, bytes) else data_json
        data = api.pad_aes_gcm(data)

        rdata = bytes(data, 'utf-8')
        rdata = crypto.encrypt_aes_v2(rdata, record_key_unencrypted)

        client_modif_time = api.current_milli_time()

        ra = record_pb2.ApplicationAddRequest()
        ra.app_uid = app_record_uid
        ra.record_key = record_key_encrypted
        ra.client_modified_time = client_modif_time
        ra.data = rdata

        api.communicate_rest(params, ra, 'vault/application_add')
        app_uid_str = utils.base64_url_encode(ra.app_uid)
        
        if format_type == 'json':
            result = {
                "app_name": app_name,
                "app_uid": app_uid_str,
                "message": "Application was successfully added",
                "created_at": datetime.datetime.fromtimestamp(client_modif_time / 1000).strftime('%Y-%m-%d %H:%M:%S')
            }
            params.sync_data = True
            return json.dumps(result, indent=2)
        else:
            print(bcolors.OKGREEN + f"Application was successfully added (UID: {app_uid_str})" + bcolors.ENDC)

        params.sync_data = True

    @staticmethod
    def remove_share(params, app_name_or_uid, secret_uids):
        app = KSMCommand.get_app_record(params, app_name_or_uid)
        if not app:
            raise Exception("KMS App with name or uid '%s' not found" % app_name_or_uid)

        app_uid = app.get('record_uid')

        rq = APIRequest_pb2.RemoveAppSharesRequest()

        rq.appRecordUid = utils.base64_url_decode(app_uid)
        rq.shares.extend((utils.base64_url_decode(x) for x in secret_uids))
        api.communicate_rest(params, rq, 'vault/app_share_remove')
        print(bcolors.OKGREEN + "Secret share was successfully removed from the application\n" + bcolors.ENDC)

    @staticmethod
    def remove_all_clients(params, app_name_or_uid, force):
        app = KSMCommand.get_app_record(params, app_name_or_uid)
        if not app:
            raise Exception("KMS App with name or uid '%s' not found" % app_name_or_uid)

        app_uid = app.get('record_uid')

        app_info = KSMCommand.get_app_info(params, app_uid)

        clients_count = len(app_info[0].clients)

        if clients_count == 0:
            print(bcolors.WARNING + "No client devices registered for this Application\n" + bcolors.ENDC)
            return

        if not force:

            print(f"This app has {clients_count} client(s) connections.")
            uc = user_choice('\tAre you sure you want to delete all clients from this application?', 'yn', default='n')
            if uc.lower() != 'y':
                return

        client_ids_to_rem = [utils.base64_url_encode(c.clientId) for ai in app_info
                             for c in ai.clients if c.appClientType == enterprise_pb2.GENERAL]
        if len(client_ids_to_rem) > 0:
            KSMCommand.remove_client(params, app_name_or_uid, client_ids_to_rem)

    @staticmethod
    def remove_client(params, app_name_or_uid, client_names_and_hashes):

        def convert_ids_and_hashes_to_hashes(cnahs, app_uid):

            client_id_hashes_bytes = []

            app_info = KSMCommand.get_app_info(params, app_uid)

            for ai in app_info:

                if len(ai.clients) > 0:
                    for c in ai.clients:
                        name = c.id
                        client_id = utils.base64_url_encode(c.clientId)

                        for cnah in cnahs:
                            if name == cnah:
                                client_id_hashes_bytes.append(c.clientId)
                            else:
                                if len(cnah) >= KSMCommand.CLIENT_SHORT_ID_LENGTH and client_id.startswith(cnah):
                                    client_id_hashes_bytes.append(c.clientId)

            return client_id_hashes_bytes

        if not app_name_or_uid:
            raise Exception("No app provided")

        app = KSMCommand.get_app_record(params, app_name_or_uid)
        if not app:
            raise Exception("KMS App with name or uid '%s' not found" % app_name_or_uid)

        app_uid = app.get('record_uid')

        client_hashes = convert_ids_and_hashes_to_hashes(client_names_and_hashes, app_uid)

        found_clients_count = len(client_hashes)
        if found_clients_count == 0:
            print(bcolors.WARNING + "No Client Devices found with given name or ID\n" + bcolors.ENDC)
            return
        else:
            uc = user_choice(f'\tAre you sure you want to delete {found_clients_count} matching clients from this application?',
                             'yn', default='n')
            if uc.lower() != 'y':
                return

        rq = APIRequest_pb2.RemoveAppClientsRequest()

        rq.appRecordUid = utils.base64_url_decode(app_uid)
        rq.clients.extend(client_hashes)
        api.communicate_rest(params, rq, 'vault/app_client_remove')
        print(bcolors.OKGREEN + "\nClient removal was successful\n" + bcolors.ENDC)

    @staticmethod
    def add_client(params, app_name_or_uid, count, unlock_ip, first_access_expire_on, access_expire_in_min,
                   client_name=None, config_init=None, silent=False, client_type=enterprise_pb2.GENERAL):

        is_ip_unlocked = as_boolean(unlock_ip, False)
        curr_ms = int(time.time() * 1000)

        first_access_expire_on_ms = curr_ms + (int(first_access_expire_on) * 60 * 1000)

        if access_expire_in_min:
            access_expire_on_ms = curr_ms + (int(access_expire_in_min) * 60 * 1000)
        else:
            access_expire_on_ms = curr_ms

        if not app_name_or_uid:
            raise Exception("No app provided")

        rec_cache_val = KSMCommand.get_app_record(params, app_name_or_uid)

        if not rec_cache_val:
            raise Exception("KMS App with name or uid '%s' not found" % app_name_or_uid)

        logging.debug("App uid=%s, unlock_ip=%s" % (rec_cache_val.get('record_uid'), unlock_ip))

        master_key = rec_cache_val.get('record_key_unencrypted')

        keys_str = ""
        otat_str = ""

        tokens = []

        for i in range(count):
            secret_bytes = os.urandom(32)
            counter_bytes = b'KEEPER_SECRETS_MANAGER_CLIENT_ID'
            digest = 'sha512'

            try:
                mac = hmac.new(secret_bytes, counter_bytes, digest).digest()
            except Exception as e:
                logging.error(e.args[0])
                return

            encrypted_master_key = crypto.encrypt_aes_v2(master_key, secret_bytes)

            rq = APIRequest_pb2.AddAppClientRequest()
            rq.appRecordUid = utils.base64_url_decode(rec_cache_val.get('record_uid'))
            rq.encryptedAppKey = encrypted_master_key
            rq.lockIp = not is_ip_unlocked
            rq.firstAccessExpireOn = first_access_expire_on_ms
            rq.appClientType = client_type

            if access_expire_in_min:
                rq.accessExpireOn = access_expire_on_ms

            rq.clientId = mac

            if client_name:
                if count == 1:
                    rq.id = client_name
                else:
                    rq.id = client_name + " " + str((i+1))

            # api_request_payload = ApiRequestPayload()
            # api_request_payload.payload = rq.SerializeToString()
            # api_request_payload.encryptedSessionToken = base64.urlsafe_b64decode(params.session_token + '==')

            # rs = execute_rest(params.rest_context, 'vault/app_client_add', api_request_payload)

            device = api.communicate_rest(params, rq, 'vault/app_client_add', rs_type=APIRequest_pb2.Device)

            encrypted_device_token = bytes_to_base64(device.encryptedDeviceToken)

            if encrypted_device_token:

                if keys_str:
                    keys_str += '\n'

                if not is_ip_unlocked:
                    lock_ip_stat = bcolors.OKGREEN + "Enabled" + bcolors.ENDC
                else:
                    lock_ip_stat = bcolors.HIGHINTENSITYRED + "Disabled" + bcolors.ENDC
                exp_date_str = bcolors.BOLD + datetime.datetime.fromtimestamp(
                    first_access_expire_on_ms / 1000).strftime('%Y-%m-%d %H:%M:%S') + bcolors.ENDC

                if access_expire_in_min:
                    app_expire_on_str = bcolors.BOLD + datetime.datetime.fromtimestamp(
                        access_expire_on_ms / 1000).strftime('%Y-%m-%d %H:%M:%S') + bcolors.ENDC
                else:
                    app_expire_on_str = bcolors.WARNING + "Never" + bcolors.ENDC

                token = utils.base64_url_encode(secret_bytes)

                if not config_init:
                    abbrev = get_abbrev_by_host(params.server.lower())

                    if abbrev:
                        token_w_prefix = f'{abbrev}:{token}'
                    else:
                        if not params.server.startswith('http'):
                            tmp_server = "https://" + params.server
                        else:
                            tmp_server = params.server

                        token_w_prefix = f'{urllib.parse.urlparse(tmp_server).netloc.lower()}:{token}'

                    otat_str += f'\nOne-Time Access Token: {bcolors.OKGREEN}{token_w_prefix}{bcolors.ENDC}\n'

                    tokens.append({
                        'oneTimeToken': token_w_prefix,
                        'deviceToken': encrypted_device_token
                    })

                else:
                    config_str = KSMCommand.init_ksm_config(params,
                                                            one_time_token=token,
                                                            config_init=config_init)
                    otat_str += f'\nInitialized Config: {bcolors.OKGREEN}{config_str}{bcolors.ENDC}\n'
                    tokens.append({
                        'config': config_str,
                        'deviceToken': encrypted_device_token
                    })

                if client_name:
                    otat_str += f'Name: {client_name}\n'

                otat_str += f'IP Lock: {lock_ip_stat}\n' \
                            f'Token Expires On: {exp_date_str}\n' \
                            f'App Access Expires on: {app_expire_on_str}\n'

        if not silent:
            print(f'\nSuccessfully generated Client Device\n'
                  f'====================================\n'
                  f'{otat_str}')

        if config_init and not unlock_ip and not silent:
            print(bcolors.WARNING + "\tWarning: Configuration is now locked to your current IP. To keep in unlock you "
                                    "can add flag `--unlock-ip` or use the One-time token to generate configuration on "
                                    "the host that has the IP that needs to be locked." + bcolors.ENDC)

            logging.warning('')

        return tokens

    @staticmethod
    def init_ksm_config(params, one_time_token, config_init, include_config_dict=False):

        try:
            from keeper_secrets_manager_core import SecretsManager
            from keeper_secrets_manager_core.configkeys import ConfigKeys
            from keeper_secrets_manager_core.storage import InMemoryKeyValueStorage
        except Exception:
            raise Exception("Keeper Secrets Manager is not installed.\n"
                            "Install it using pip `pip3 install keeper-secrets-manager-core`")

        ksm_conf_storage = InMemoryKeyValueStorage()

        secrets_manager = SecretsManager(
            hostname=params.config.get('server'),
            token=one_time_token,
            # verify_ssl_certs=False,
            config=ksm_conf_storage
        )

        secrets_manager.get_secrets("NON-EXISTING-RECORD-UID")

        config_dict = {
            ConfigKeys.KEY_HOSTNAME.value:             ksm_conf_storage.config.get(ConfigKeys.KEY_HOSTNAME),
            ConfigKeys.KEY_CLIENT_ID.value:            ksm_conf_storage.config.get(ConfigKeys.KEY_CLIENT_ID),
            ConfigKeys.KEY_PRIVATE_KEY.value:          ksm_conf_storage.config.get(ConfigKeys.KEY_PRIVATE_KEY),
            ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID.value: ksm_conf_storage.config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID),
            ConfigKeys.KEY_APP_KEY.value:              ksm_conf_storage.config.get(ConfigKeys.KEY_APP_KEY),
        }

        # if the SDK version is below 16.2.0 then this key won't be present
        if 'KEY_OWNER_PUBLIC_KEY' in ConfigKeys.__members__ and ksm_conf_storage.config.get(ConfigKeys.KEY_OWNER_PUBLIC_KEY):
            config_dict[ConfigKeys.KEY_OWNER_PUBLIC_KEY.value] = ksm_conf_storage.config.get(ConfigKeys.KEY_OWNER_PUBLIC_KEY)

        converted_config = KSMCommand.convert_config_dict(config_dict, config_init)

        if include_config_dict:
            return {
                'config_str': converted_config,
                'config_dict': config_dict
            }
        else:
            return converted_config

    @staticmethod
    def convert_config_dict(config_dict, conversion_type='json'):

        config = json.dumps(config_dict)

        if conversion_type in ['b64', 'k8s']:
            config = json_to_base64(config)

        if conversion_type == 'k8s':
            config = "\n" \
                     + "apiVersion: v1\n" \
                     + "data:\n" \
                     + "  config: " + config + "\n" \
                     + "kind: Secret\n" \
                     + "metadata:\n" \
                     + "  name: ksm-config\n" \
                     + "  namespace: default\n" \
                     + "type: Opaque"

        if conversion_type == 'dict':
            config = config_dict

        return config

    @staticmethod
    def get_hash_of_one_time_token(one_time_token):
        """ KSM: Get client ID from one time token, which is equal to a Hash of one time token"""

        ott_parts = one_time_token.split(":")

        if len(ott_parts) == 2:
            ott = ott_parts[1]
        else:
            ott = ott_parts[0]

        existing_secret_key_bytes = url_safe_str_to_bytes(ott)
        digest = 'sha512'
        one_time_token_hash = bytes_to_base64(hmac.new(existing_secret_key_bytes,
                                                       b'KEEPER_SECRETS_MANAGER_CLIENT_ID',
                                                       digest).digest())
        return one_time_token_hash





import base64
import json
import os
from typing import List, Union, Optional, Dict
from unittest import mock

from Cryptodome.Cipher import AES
from Cryptodome.PublicKey import RSA

from keepercommander import api, params, shared_folder, team, crypto, utils, vault, vault_extensions, record_facades
from keepercommander.proto import record_pb2, SyncDown_pb2

_USER_NAME = 'unit.test@company.com'
_USER_PASSWORD = base64.b64encode(os.urandom(8)).decode('utf-8').strip('=')
_USER_ITERATIONS = 1000
_USER_SALT = os.urandom(16)
_USER_DATA_KEY = utils.generate_aes_key()
_USER_ACCOUNT_UID = utils.generate_uid()

_SESSION_TOKEN = base64.urlsafe_b64encode(os.urandom(64)).decode('utf-8').strip('=')

_2FA_ONE_TIME_TOKEN = '123456'
_2FA_DEVICE_TOKEN = base64.urlsafe_b64encode(os.urandom(32)).decode('utf-8').strip('=')

_private_key, _public_key = crypto.generate_rsa_key()

_DER_PRIVATE_KEY = crypto.unload_rsa_private_key(_private_key)
_ENCRYPTED_PRIVATE_KEY = utils.base64_url_encode(crypto.encrypt_aes_v1(_DER_PRIVATE_KEY, _USER_DATA_KEY))

_IMPORTED_PUBLIC_KEY = crypto.unload_rsa_public_key(_public_key)

_V2_DERIVED_KEY = crypto.derive_keyhash_v2('data_key', _USER_PASSWORD, _USER_SALT, _USER_ITERATIONS)
_dk = crypto.encrypt_aes_v2(_USER_DATA_KEY, _V2_DERIVED_KEY)
_ENCRYPTED_DATA_KEY = base64.urlsafe_b64encode(_dk).decode('utf-8').strip()

_V1_DERIVED_KEY = crypto.derive_keyhash_v1(_USER_PASSWORD, _USER_SALT, _USER_ITERATIONS)
_enc_iter = int.to_bytes(_USER_ITERATIONS, length=3, byteorder='big', signed=False)
_enc_iv = os.urandom(16)
_cipher = AES.new(_V1_DERIVED_KEY, AES.MODE_CBC, _enc_iv)
_enc_dk = b'\x01' + _enc_iter + _USER_SALT + _enc_iv + _cipher.encrypt(_USER_DATA_KEY + _USER_DATA_KEY)
_ENCRYPTION_PARAMS = base64.urlsafe_b64encode(_enc_dk).decode('utf-8').strip('=')


class VaultEnvironment:
    def __init__(self):
        self.user = _USER_NAME
        self.password = _USER_PASSWORD
        self.iterations = _USER_ITERATIONS
        self.salt = _USER_SALT
        self.data_key = _USER_DATA_KEY
        self.public_key = _public_key
        self.encoded_public_key = utils.base64_url_encode(_IMPORTED_PUBLIC_KEY)
        self.session_token = _SESSION_TOKEN
        self.one_time_token = _2FA_ONE_TIME_TOKEN
        self.device_token = _2FA_DEVICE_TOKEN
        self.encrypted_private_key = _ENCRYPTED_PRIVATE_KEY
        self.encrypted_data_key = _ENCRYPTED_DATA_KEY
        self.encryption_params = _ENCRYPTION_PARAMS
        self.revision = _REVISION


def get_user_params():
    p = params.KeeperParams(server='https://test.keepersecurity.com/')
    p.user = _USER_NAME
    p.password = _USER_PASSWORD
    return p


def get_connected_params():
    p = get_user_params()
    p.iterations = _USER_ITERATIONS
    p.salt = _USER_SALT
    p.data_key = _USER_DATA_KEY
    p.account_uid_bytes = utils.base64_url_decode(_USER_ACCOUNT_UID)

    p.auth_verifier = utils.base64_url_encode(utils.create_auth_verifier(_USER_PASSWORD, _USER_SALT, _USER_ITERATIONS))
    p.rsa_key = RSA.importKey(_DER_PRIVATE_KEY)
    p.rsa_key2 = crypto.load_rsa_private_key(_DER_PRIVATE_KEY)
    p.session_token = _SESSION_TOKEN
    return p


def get_sync_down_responses(p, request, endpoint, rs_type):
    if endpoint == 'vault/sync_down':
        return get_sync_down_response()
    elif endpoint == 'vault/get_record_types':
        content = json.dumps({
            "$id": "login",
            "categories": ["login"],
            "description": "Login template",
            "fields": [
                {"$ref": "login"},
                {"$ref": "password"},
                {"$ref": "url"},
                {"$ref": "fileRef"},
                {"$ref": "oneTimeCode"}
            ]
        })
        rs = record_pb2.RecordTypesResponse()
        rt = record_pb2.RecordType()
        rt.recordTypeId = 1
        rt.content = content
        rt.scope = record_pb2.RT_STANDARD
        rs.recordTypes.append(rt)
        rs.standardCounter = 1
        return rs
    raise NotImplementedError()


def get_synced_params():
    p = get_connected_params()
    with mock.patch('keepercommander.api.communicate_rest') as mock_comm:
        mock_comm.side_effect = get_sync_down_responses
        api.sync_down(p)
    return p


_REVISION = 100
_RECORDS = []                     # type: List[SyncDown_pb2.Record]
_RECORD_METADATA = []             # type: List[SyncDown_pb2.RecordMetaData]
_SHARED_FOLDERS = []              # type: List[SyncDown_pb2.SharedFolder]
_SHARED_FOLDER_USERS = []         # type: List[SyncDown_pb2.SharedFolderUser]
_SHARED_FOLDER_TEAMS = []         # type: List[SyncDown_pb2.SharedFolderTeam]
_SHARED_FOLDER_RECORDS = []       # type: List[SyncDown_pb2.SharedFolderRecord]
_USER_FOLDERS = []                # type: List[SyncDown_pb2.UserFolder]
_USER_FOLDER_RECORDS = []         # type: List[SyncDown_pb2.UserFolderRecord]
_USER_FOLDER_SHARED_FOLDER = []   # type: List[SyncDown_pb2.UserFolderSharedFolder]
_TEAMS = []                       # type: List[SyncDown_pb2.Team]


def get_sync_down_response():    # type: () -> SyncDown_pb2.SyncDownResponse
    response = SyncDown_pb2.SyncDownResponse()
    response.continuationToken = crypto.get_random_bytes(64)
    response.hasMore = False
    response.cacheStatus = SyncDown_pb2.CLEAR
    response.teams.extend(_TEAMS)
    response.userFolders.extend(_USER_FOLDERS)
    response.sharedFolders.extend(_SHARED_FOLDERS)
    response.sharedFolderUsers.extend(_SHARED_FOLDER_USERS)
    response.sharedFolderTeams.extend(_SHARED_FOLDER_TEAMS)
    response.sharedFolderRecords.extend(_SHARED_FOLDER_RECORDS)
    response.records.extend(_RECORDS)
    response.recordMetaData.extend(_RECORD_METADATA)
    response.userFolderRecords.extend(_USER_FOLDER_RECORDS)
    response.userFolderSharedFolders.extend(_USER_FOLDER_SHARED_FOLDER)

    return response


def register_record(rec, key_type=None):
    # type: (Union[vault.PasswordRecord, vault.TypedRecord], Optional[int]) -> bytes

    extra = None   # type: Optional[dict]
    udata = None   # type: Optional[dict]
    if isinstance(rec, vault.PasswordRecord):
        data = vault_extensions.extract_password_record_data(rec)
        extra = vault_extensions.extract_password_record_extras(rec, None)
        if rec.attachments:
            udata = {'file_id': [x.id for x in rec.attachments]}
    elif isinstance(rec, vault.TypedRecord):
        data = vault_extensions.extract_typed_record_data(rec)
    else:
        raise Exception('Unsupported record type')

    record_key = utils.generate_aes_key() if key_type != 0 else _USER_DATA_KEY

    rec_object = SyncDown_pb2.Record()
    rec_object.recordUid = utils.base64_url_decode(rec.record_uid)
    rec_object.revision = rec.revision if (0 < rec.revision <= _REVISION) else _REVISION
    rec_object.version = 2 if isinstance(rec, vault.PasswordRecord) else 3
    rec_object.shared = key_type not in [0, 1]
    rec_object.clientModifiedTime = utils.current_milli_time()
    if isinstance(rec, vault.PasswordRecord):
        rec_object.data = crypto.encrypt_aes_v1(json.dumps(data).encode('utf-8'), record_key)
        if extra:
            rec_object.extra = crypto.encrypt_aes_v1(json.dumps(extra).encode('utf-8'), record_key)
        if udata:
            rec_object.udata = json.dumps(udata)
    else:
        rec_object.data = crypto.encrypt_aes_v2(json.dumps(data).encode('utf-8'), record_key)

    _RECORDS.append(rec_object)

    if isinstance(key_type, int):
        meta_data = SyncDown_pb2.RecordMetaData()
        meta_data.recordUid = utils.base64_url_decode(rec.record_uid)
        meta_data.owner = key_type in [0, 1]
        meta_data.recordKeyType = key_type
        meta_data.canShare = key_type == 1
        meta_data.canEdit = key_type == 1

        if key_type == 0:
            pass
        if key_type == 1:
            meta_data.recordKey = crypto.encrypt_aes_v1(record_key, _USER_DATA_KEY)
        elif key_type == 2:
            meta_data.recordKey = crypto.encrypt_rsa(record_key, _public_key)

        _RECORD_METADATA.append(meta_data)

    return record_key


def register_records_to_folder(folder_uid, record_uids):    # type: (Optional[str], list) -> None
    for record_uid in record_uids:
        ufr = SyncDown_pb2.UserFolderRecord()
        ufr.recordUid = utils.base64_url_decode(record_uid)
        if folder_uid:
            ufr.folderUid = utils.base64_url_decode(folder_uid)
        _USER_FOLDER_RECORDS.append(ufr)


def register_shared_folder(shared_folder, records):  # type: (shared_folder.SharedFolder, Dict[str, bytes]) -> bytes
    shared_folder_key = utils.generate_aes_key()

    sf = SyncDown_pb2.SharedFolder()
    sf.sharedFolderUid = utils.base64_url_decode(shared_folder.shared_folder_uid)
    sf.revision = 5
    sf.sharedFolderKey = crypto.encrypt_aes_v1(shared_folder_key, _USER_DATA_KEY)
    sf.keyType = 1
    sf.data =  crypto.encrypt_aes_v1(json.dumps({'name': shared_folder.name}).encode(), shared_folder_key)
    sf.defaultManageRecords = True
    sf.defaultManageUsers = True
    sf.defaultCanEdit = True
    sf.defaultCanReshare = True
    sf.cacheStatus = SyncDown_pb2.CLEAR
    sf.name = crypto.encrypt_aes_v1(shared_folder.name.encode('utf-8'), shared_folder_key)
    _SHARED_FOLDERS.append(sf)

    sfu = SyncDown_pb2.SharedFolderUser()
    sfu.sharedFolderUid = utils.base64_url_decode(shared_folder.shared_folder_uid)
    sfu.manageRecords = True
    sfu.manageUsers = True
    sfu.accountUid = utils.base64_url_decode(_USER_ACCOUNT_UID)
    _SHARED_FOLDER_USERS.append(sfu)

    for record_uid, record_key in records.items():
        sfr = SyncDown_pb2.SharedFolderRecord()
        sfr.sharedFolderUid = utils.base64_url_decode(shared_folder.shared_folder_uid)
        sfr.recordUid = utils.base64_url_decode(record_uid)
        sfr.recordKey = crypto.encrypt_aes_v1(record_key, shared_folder_key)
        sfr.canShare = False
        sfr.canEdit = False
        sfr.ownerAccountUid = utils.base64_url_decode(_USER_ACCOUNT_UID)
        sfr.owner = True
        _SHARED_FOLDER_RECORDS.append(sfr)

    return shared_folder_key


def register_team(team, key_type, sfs=None):     # type: (team.Team, int, dict) -> bytes
    public_key = crypto.load_rsa_public_key(_IMPORTED_PUBLIC_KEY)
    team_key = utils.generate_aes_key()

    t = SyncDown_pb2.Team()
    t.teamUid = utils.base64_url_decode(team.team_uid)
    t.name = team.name
    t.teamKey = crypto.encrypt_aes_v1(team_key, _USER_DATA_KEY) \
        if key_type == 1 else crypto.encrypt_rsa(team_key, public_key)
    t.teamKeyType = key_type
    t.restrictEdit = team.restrict_edit
    t.restrictShare = team.restrict_share
    t.restrictView = team.restrict_view
    _TEAMS.append(t)

    if isinstance(sfs, dict):
        for shared_folder_uid, shared_folder_key in sfs.items():
            sfk = SyncDown_pb2.SharedFolderKey()
            sfk.sharedFolderUid = utils.base64_url_decode(shared_folder_uid)
            sfk.sharedFolderKey = crypto.encrypt_aes_v1(shared_folder_key, team_key)
            t.sharedFolderKeys.append(sfk)

            sft = SyncDown_pb2.SharedFolderTeam()
            sft.sharedFolderUid = utils.base64_url_decode(shared_folder_uid)
            sft.teamUid = utils.base64_url_decode(team.team_uid)
            sft.name = team.name
            sft.manageRecords = key_type == 1
            sft.manageUsers = key_type == 1
            _SHARED_FOLDER_TEAMS.append(sft)

    return team_key


def generate_data():
    r1 = vault.PasswordRecord()
    r1.record_uid = utils.generate_uid()
    r1.title = 'Record 1'
    r1.login = 'user1@keepersecurity.com'
    r1.password = 'password1'
    r1.login_url = 'https://keepersecurity.com/1'
    r1.set_custom_value('field1', 'value1')
    r1.notes = 'note1'
    atta1 = vault.AttachmentFile()
    atta1.id = 'ABCDEFGH'
    atta1.name = 'Attachment 1'
    atta1.key = utils.base64_url_encode(api.generate_aes_key())
    atta1.size = 1000
    r1.attachments = [atta1]
    r1.revision = 1
    r1_key = register_record(r1, 1)

    facade = record_facades.LoginRecordFacade()
    r2 = vault.TypedRecord()
    r2.record_uid = utils.generate_uid()
    r2.title = 'Record 2'
    r2.revision = 2
    facade.record = r2
    facade.login = 'user2@keepersecurity.com'
    facade.password = 'password2'
    facade.login_url = 'https://keepersecurity.com/2'
    facade.notes = 'note2'
    r2.custom.append(vault.TypedField.new_field('text', 'value2', 'field2'))
    r2_key = register_record(r2, 2)

    register_records_to_folder(None, [r1.record_uid, r2.record_uid])

    r3 = vault.PasswordRecord()
    r3.record_uid = utils.generate_uid()
    r3.title = 'Record 3'
    r3.login = 'user3@keepersecurity.com'
    r3.password = 'password3'
    r3.login_url = 'https://keepersecurity.com/3'
    r3.revision = 3
    r3_key = register_record(r3)

    sf1 = shared_folder.SharedFolder()
    sf1.shared_folder_uid = utils.generate_uid()
    sf1.default_manage_records = False
    sf1.default_manage_users = False
    sf1.default_can_edit = False
    sf1.default_can_share = False
    sf1.name = 'Shared Folder 1'
    sf1_key = register_shared_folder(sf1, {
        r3.record_uid: r3_key
    })
    register_records_to_folder(sf1.shared_folder_uid, [r3.record_uid])
    ufsf = SyncDown_pb2.UserFolderSharedFolder()
    ufsf.sharedFolderUid = utils.base64_url_decode(sf1.shared_folder_uid)
    _USER_FOLDER_SHARED_FOLDER.append(ufsf)

    t1 = team.Team()
    t1.team_uid = utils.generate_uid()
    t1.name = 'Team 1'
    t1.restrict_edit = True
    t1.restrict_share = True
    t1.restrict_view = False

    register_team(t1, 1, {sf1.shared_folder_uid: sf1_key})

    folder_key = utils.generate_aes_key()
    uf = SyncDown_pb2.UserFolder()
    uf.folderUid = utils.base64_url_decode(utils.generate_uid())
    uf.userFolderKey = crypto.encrypt_aes_v1(folder_key, _USER_DATA_KEY)
    uf.keyType = 1
    uf.revision = 4
    uf.data = crypto.encrypt_aes_v1(json.dumps({'name': 'User Folder 1'}).encode('utf-8'), folder_key)
    _USER_FOLDERS.append(uf)

generate_data()



import enterprise_pb2 as _enterprise_pb2
import record_pb2 as _record_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class WebRtcConnectionType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    CONNECTION: _ClassVar[WebRtcConnectionType]
    TUNNEL: _ClassVar[WebRtcConnectionType]
    SSH: _ClassVar[WebRtcConnectionType]
    RDP: _ClassVar[WebRtcConnectionType]
    HTTP: _ClassVar[WebRtcConnectionType]
    VNC: _ClassVar[WebRtcConnectionType]
    TELNET: _ClassVar[WebRtcConnectionType]
    MYSQL: _ClassVar[WebRtcConnectionType]
    SQL_SERVER: _ClassVar[WebRtcConnectionType]
    POSTGRESQL: _ClassVar[WebRtcConnectionType]

class PAMOperationType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    ADD: _ClassVar[PAMOperationType]
    UPDATE: _ClassVar[PAMOperationType]
    REPLACE: _ClassVar[PAMOperationType]
    DELETE: _ClassVar[PAMOperationType]

class PAMOperationResultType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    POT_SUCCESS: _ClassVar[PAMOperationResultType]
    POT_UNKNOWN_ERROR: _ClassVar[PAMOperationResultType]
    POT_ALREADY_EXISTS: _ClassVar[PAMOperationResultType]
    POT_DOES_NOT_EXIST: _ClassVar[PAMOperationResultType]

class ControllerMessageType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    CMT_GENERAL: _ClassVar[ControllerMessageType]
    CMT_ROTATE: _ClassVar[ControllerMessageType]
    CMT_DISCOVERY: _ClassVar[ControllerMessageType]
    CMT_CONNECT: _ClassVar[ControllerMessageType]

class PAMRecordingType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    PRT_SESSION: _ClassVar[PAMRecordingType]
    PRT_TYPESCRIPT: _ClassVar[PAMRecordingType]
    PRT_TIME: _ClassVar[PAMRecordingType]
CONNECTION: WebRtcConnectionType
TUNNEL: WebRtcConnectionType
SSH: WebRtcConnectionType
RDP: WebRtcConnectionType
HTTP: WebRtcConnectionType
VNC: WebRtcConnectionType
TELNET: WebRtcConnectionType
MYSQL: WebRtcConnectionType
SQL_SERVER: WebRtcConnectionType
POSTGRESQL: WebRtcConnectionType
ADD: PAMOperationType
UPDATE: PAMOperationType
REPLACE: PAMOperationType
DELETE: PAMOperationType
POT_SUCCESS: PAMOperationResultType
POT_UNKNOWN_ERROR: PAMOperationResultType
POT_ALREADY_EXISTS: PAMOperationResultType
POT_DOES_NOT_EXIST: PAMOperationResultType
CMT_GENERAL: ControllerMessageType
CMT_ROTATE: ControllerMessageType
CMT_DISCOVERY: ControllerMessageType
CMT_CONNECT: ControllerMessageType
PRT_SESSION: PAMRecordingType
PRT_TYPESCRIPT: PAMRecordingType
PRT_TIME: PAMRecordingType

class PAMRotationSchedule(_message.Message):
    __slots__ = ["recordUid", "configurationUid", "controllerUid", "scheduleData", "noSchedule"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    CONFIGURATIONUID_FIELD_NUMBER: _ClassVar[int]
    CONTROLLERUID_FIELD_NUMBER: _ClassVar[int]
    SCHEDULEDATA_FIELD_NUMBER: _ClassVar[int]
    NOSCHEDULE_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    configurationUid: bytes
    controllerUid: bytes
    scheduleData: str
    noSchedule: bool
    def __init__(self, recordUid: _Optional[bytes] = ..., configurationUid: _Optional[bytes] = ..., controllerUid: _Optional[bytes] = ..., scheduleData: _Optional[str] = ..., noSchedule: bool = ...) -> None: ...

class PAMRotationSchedulesResponse(_message.Message):
    __slots__ = ["schedules"]
    SCHEDULES_FIELD_NUMBER: _ClassVar[int]
    schedules: _containers.RepeatedCompositeFieldContainer[PAMRotationSchedule]
    def __init__(self, schedules: _Optional[_Iterable[_Union[PAMRotationSchedule, _Mapping]]] = ...) -> None: ...

class PAMOnlineController(_message.Message):
    __slots__ = ["controllerUid", "connectedOn", "ipAddress", "version", "connections"]
    CONTROLLERUID_FIELD_NUMBER: _ClassVar[int]
    CONNECTEDON_FIELD_NUMBER: _ClassVar[int]
    IPADDRESS_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    CONNECTIONS_FIELD_NUMBER: _ClassVar[int]
    controllerUid: bytes
    connectedOn: int
    ipAddress: str
    version: str
    connections: _containers.RepeatedCompositeFieldContainer[PAMWebRtcConnection]
    def __init__(self, controllerUid: _Optional[bytes] = ..., connectedOn: _Optional[int] = ..., ipAddress: _Optional[str] = ..., version: _Optional[str] = ..., connections: _Optional[_Iterable[_Union[PAMWebRtcConnection, _Mapping]]] = ...) -> None: ...

class PAMWebRtcConnection(_message.Message):
    __slots__ = ["connectionUid", "type", "recordUid", "userName", "startedOn", "configurationUid"]
    CONNECTIONUID_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    STARTEDON_FIELD_NUMBER: _ClassVar[int]
    CONFIGURATIONUID_FIELD_NUMBER: _ClassVar[int]
    connectionUid: bytes
    type: WebRtcConnectionType
    recordUid: bytes
    userName: str
    startedOn: int
    configurationUid: bytes
    def __init__(self, connectionUid: _Optional[bytes] = ..., type: _Optional[_Union[WebRtcConnectionType, str]] = ..., recordUid: _Optional[bytes] = ..., userName: _Optional[str] = ..., startedOn: _Optional[int] = ..., configurationUid: _Optional[bytes] = ...) -> None: ...

class PAMOnlineControllers(_message.Message):
    __slots__ = ["deprecated", "controllers"]
    DEPRECATED_FIELD_NUMBER: _ClassVar[int]
    CONTROLLERS_FIELD_NUMBER: _ClassVar[int]
    deprecated: _containers.RepeatedScalarFieldContainer[bytes]
    controllers: _containers.RepeatedCompositeFieldContainer[PAMOnlineController]
    def __init__(self, deprecated: _Optional[_Iterable[bytes]] = ..., controllers: _Optional[_Iterable[_Union[PAMOnlineController, _Mapping]]] = ...) -> None: ...

class PAMRotateRequest(_message.Message):
    __slots__ = ["requestUid", "recordUid"]
    REQUESTUID_FIELD_NUMBER: _ClassVar[int]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    requestUid: bytes
    recordUid: bytes
    def __init__(self, requestUid: _Optional[bytes] = ..., recordUid: _Optional[bytes] = ...) -> None: ...

class PAMControllersResponse(_message.Message):
    __slots__ = ["controllers"]
    CONTROLLERS_FIELD_NUMBER: _ClassVar[int]
    controllers: _containers.RepeatedCompositeFieldContainer[PAMController]
    def __init__(self, controllers: _Optional[_Iterable[_Union[PAMController, _Mapping]]] = ...) -> None: ...

class PAMRemoveController(_message.Message):
    __slots__ = ["controllerUid", "message"]
    CONTROLLERUID_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    controllerUid: bytes
    message: str
    def __init__(self, controllerUid: _Optional[bytes] = ..., message: _Optional[str] = ...) -> None: ...

class PAMRemoveControllerResponse(_message.Message):
    __slots__ = ["controllers"]
    CONTROLLERS_FIELD_NUMBER: _ClassVar[int]
    controllers: _containers.RepeatedCompositeFieldContainer[PAMRemoveController]
    def __init__(self, controllers: _Optional[_Iterable[_Union[PAMRemoveController, _Mapping]]] = ...) -> None: ...

class PAMModifyRequest(_message.Message):
    __slots__ = ["operations"]
    OPERATIONS_FIELD_NUMBER: _ClassVar[int]
    operations: _containers.RepeatedCompositeFieldContainer[PAMDataOperation]
    def __init__(self, operations: _Optional[_Iterable[_Union[PAMDataOperation, _Mapping]]] = ...) -> None: ...

class PAMDataOperation(_message.Message):
    __slots__ = ["operationType", "configuration", "element"]
    OPERATIONTYPE_FIELD_NUMBER: _ClassVar[int]
    CONFIGURATION_FIELD_NUMBER: _ClassVar[int]
    ELEMENT_FIELD_NUMBER: _ClassVar[int]
    operationType: PAMOperationType
    configuration: PAMConfigurationData
    element: PAMElementData
    def __init__(self, operationType: _Optional[_Union[PAMOperationType, str]] = ..., configuration: _Optional[_Union[PAMConfigurationData, _Mapping]] = ..., element: _Optional[_Union[PAMElementData, _Mapping]] = ...) -> None: ...

class PAMConfigurationData(_message.Message):
    __slots__ = ["configurationUid", "nodeId", "controllerUid", "data"]
    CONFIGURATIONUID_FIELD_NUMBER: _ClassVar[int]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    CONTROLLERUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    configurationUid: bytes
    nodeId: int
    controllerUid: bytes
    data: bytes
    def __init__(self, configurationUid: _Optional[bytes] = ..., nodeId: _Optional[int] = ..., controllerUid: _Optional[bytes] = ..., data: _Optional[bytes] = ...) -> None: ...

class PAMElementData(_message.Message):
    __slots__ = ["elementUid", "parentUid", "data"]
    ELEMENTUID_FIELD_NUMBER: _ClassVar[int]
    PARENTUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    elementUid: bytes
    parentUid: bytes
    data: bytes
    def __init__(self, elementUid: _Optional[bytes] = ..., parentUid: _Optional[bytes] = ..., data: _Optional[bytes] = ...) -> None: ...

class PAMElementOperationResult(_message.Message):
    __slots__ = ["elementUid", "result", "message"]
    ELEMENTUID_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    elementUid: bytes
    result: PAMOperationResultType
    message: str
    def __init__(self, elementUid: _Optional[bytes] = ..., result: _Optional[_Union[PAMOperationResultType, str]] = ..., message: _Optional[str] = ...) -> None: ...

class PAMModifyResult(_message.Message):
    __slots__ = ["results"]
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    results: _containers.RepeatedCompositeFieldContainer[PAMElementOperationResult]
    def __init__(self, results: _Optional[_Iterable[_Union[PAMElementOperationResult, _Mapping]]] = ...) -> None: ...

class PAMElement(_message.Message):
    __slots__ = ["elementUid", "data", "created", "lastModified", "children"]
    ELEMENTUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    CREATED_FIELD_NUMBER: _ClassVar[int]
    LASTMODIFIED_FIELD_NUMBER: _ClassVar[int]
    CHILDREN_FIELD_NUMBER: _ClassVar[int]
    elementUid: bytes
    data: bytes
    created: int
    lastModified: int
    children: _containers.RepeatedCompositeFieldContainer[PAMElement]
    def __init__(self, elementUid: _Optional[bytes] = ..., data: _Optional[bytes] = ..., created: _Optional[int] = ..., lastModified: _Optional[int] = ..., children: _Optional[_Iterable[_Union[PAMElement, _Mapping]]] = ...) -> None: ...

class PAMGenericUidRequest(_message.Message):
    __slots__ = ["uid"]
    UID_FIELD_NUMBER: _ClassVar[int]
    uid: bytes
    def __init__(self, uid: _Optional[bytes] = ...) -> None: ...

class PAMGenericUidsRequest(_message.Message):
    __slots__ = ["uids"]
    UIDS_FIELD_NUMBER: _ClassVar[int]
    uids: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, uids: _Optional[_Iterable[bytes]] = ...) -> None: ...

class PAMConfiguration(_message.Message):
    __slots__ = ["configurationUid", "nodeId", "controllerUid", "data", "created", "lastModified", "children"]
    CONFIGURATIONUID_FIELD_NUMBER: _ClassVar[int]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    CONTROLLERUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    CREATED_FIELD_NUMBER: _ClassVar[int]
    LASTMODIFIED_FIELD_NUMBER: _ClassVar[int]
    CHILDREN_FIELD_NUMBER: _ClassVar[int]
    configurationUid: bytes
    nodeId: int
    controllerUid: bytes
    data: bytes
    created: int
    lastModified: int
    children: _containers.RepeatedCompositeFieldContainer[PAMElement]
    def __init__(self, configurationUid: _Optional[bytes] = ..., nodeId: _Optional[int] = ..., controllerUid: _Optional[bytes] = ..., data: _Optional[bytes] = ..., created: _Optional[int] = ..., lastModified: _Optional[int] = ..., children: _Optional[_Iterable[_Union[PAMElement, _Mapping]]] = ...) -> None: ...

class PAMConfigurations(_message.Message):
    __slots__ = ["configurations"]
    CONFIGURATIONS_FIELD_NUMBER: _ClassVar[int]
    configurations: _containers.RepeatedCompositeFieldContainer[PAMConfiguration]
    def __init__(self, configurations: _Optional[_Iterable[_Union[PAMConfiguration, _Mapping]]] = ...) -> None: ...

class PAMController(_message.Message):
    __slots__ = ["controllerUid", "controllerName", "deviceToken", "deviceName", "nodeId", "created", "lastModified", "applicationUid", "appClientType", "isInitialized"]
    CONTROLLERUID_FIELD_NUMBER: _ClassVar[int]
    CONTROLLERNAME_FIELD_NUMBER: _ClassVar[int]
    DEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    DEVICENAME_FIELD_NUMBER: _ClassVar[int]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    CREATED_FIELD_NUMBER: _ClassVar[int]
    LASTMODIFIED_FIELD_NUMBER: _ClassVar[int]
    APPLICATIONUID_FIELD_NUMBER: _ClassVar[int]
    APPCLIENTTYPE_FIELD_NUMBER: _ClassVar[int]
    ISINITIALIZED_FIELD_NUMBER: _ClassVar[int]
    controllerUid: bytes
    controllerName: str
    deviceToken: str
    deviceName: str
    nodeId: int
    created: int
    lastModified: int
    applicationUid: bytes
    appClientType: _enterprise_pb2.AppClientType
    isInitialized: bool
    def __init__(self, controllerUid: _Optional[bytes] = ..., controllerName: _Optional[str] = ..., deviceToken: _Optional[str] = ..., deviceName: _Optional[str] = ..., nodeId: _Optional[int] = ..., created: _Optional[int] = ..., lastModified: _Optional[int] = ..., applicationUid: _Optional[bytes] = ..., appClientType: _Optional[_Union[_enterprise_pb2.AppClientType, str]] = ..., isInitialized: bool = ...) -> None: ...

class ControllerResponse(_message.Message):
    __slots__ = ["payload"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: str
    def __init__(self, payload: _Optional[str] = ...) -> None: ...

class PAMConfigurationController(_message.Message):
    __slots__ = ["configurationUid", "controllerUid"]
    CONFIGURATIONUID_FIELD_NUMBER: _ClassVar[int]
    CONTROLLERUID_FIELD_NUMBER: _ClassVar[int]
    configurationUid: bytes
    controllerUid: bytes
    def __init__(self, configurationUid: _Optional[bytes] = ..., controllerUid: _Optional[bytes] = ...) -> None: ...

class ConfigurationAddRequest(_message.Message):
    __slots__ = ["configurationUid", "recordKey", "data", "recordLinks", "audit"]
    CONFIGURATIONUID_FIELD_NUMBER: _ClassVar[int]
    RECORDKEY_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    RECORDLINKS_FIELD_NUMBER: _ClassVar[int]
    AUDIT_FIELD_NUMBER: _ClassVar[int]
    configurationUid: bytes
    recordKey: bytes
    data: bytes
    recordLinks: _containers.RepeatedCompositeFieldContainer[_record_pb2.RecordLink]
    audit: _record_pb2.RecordAudit
    def __init__(self, configurationUid: _Optional[bytes] = ..., recordKey: _Optional[bytes] = ..., data: _Optional[bytes] = ..., recordLinks: _Optional[_Iterable[_Union[_record_pb2.RecordLink, _Mapping]]] = ..., audit: _Optional[_Union[_record_pb2.RecordAudit, _Mapping]] = ...) -> None: ...

class RelayAccessCreds(_message.Message):
    __slots__ = ["username", "password", "serverTime"]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    SERVERTIME_FIELD_NUMBER: _ClassVar[int]
    username: str
    password: str
    serverTime: int
    def __init__(self, username: _Optional[str] = ..., password: _Optional[str] = ..., serverTime: _Optional[int] = ...) -> None: ...

class PAMRecording(_message.Message):
    __slots__ = ["connectionUid", "recordingType", "recordUid", "userName", "startedOn", "length", "fileSize", "protocol"]
    CONNECTIONUID_FIELD_NUMBER: _ClassVar[int]
    RECORDINGTYPE_FIELD_NUMBER: _ClassVar[int]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    STARTEDON_FIELD_NUMBER: _ClassVar[int]
    LENGTH_FIELD_NUMBER: _ClassVar[int]
    FILESIZE_FIELD_NUMBER: _ClassVar[int]
    PROTOCOL_FIELD_NUMBER: _ClassVar[int]
    connectionUid: bytes
    recordingType: PAMRecordingType
    recordUid: bytes
    userName: str
    startedOn: int
    length: int
    fileSize: int
    protocol: str
    def __init__(self, connectionUid: _Optional[bytes] = ..., recordingType: _Optional[_Union[PAMRecordingType, str]] = ..., recordUid: _Optional[bytes] = ..., userName: _Optional[str] = ..., startedOn: _Optional[int] = ..., length: _Optional[int] = ..., fileSize: _Optional[int] = ..., protocol: _Optional[str] = ...) -> None: ...

class PAMRecordingsResponse(_message.Message):
    __slots__ = ["recordings"]
    RECORDINGS_FIELD_NUMBER: _ClassVar[int]
    recordings: _containers.RepeatedCompositeFieldContainer[PAMRecording]
    def __init__(self, recordings: _Optional[_Iterable[_Union[PAMRecording, _Mapping]]] = ...) -> None: ...

class PAMData(_message.Message):
    __slots__ = ["vertex", "content"]
    VERTEX_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    vertex: bytes
    content: bytes
    def __init__(self, vertex: _Optional[bytes] = ..., content: _Optional[bytes] = ...) -> None: ...

class UidList(_message.Message):
    __slots__ = ["uids"]
    UIDS_FIELD_NUMBER: _ClassVar[int]
    uids: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, uids: _Optional[_Iterable[bytes]] = ...) -> None: ...

class PAMResourceConfig(_message.Message):
    __slots__ = ["recordUid", "networkUid", "adminUid", "meta", "connectionSettings", "connectUsers", "domainUid", "jitSettings"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    NETWORKUID_FIELD_NUMBER: _ClassVar[int]
    ADMINUID_FIELD_NUMBER: _ClassVar[int]
    META_FIELD_NUMBER: _ClassVar[int]
    CONNECTIONSETTINGS_FIELD_NUMBER: _ClassVar[int]
    CONNECTUSERS_FIELD_NUMBER: _ClassVar[int]
    DOMAINUID_FIELD_NUMBER: _ClassVar[int]
    JITSETTINGS_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    networkUid: bytes
    adminUid: bytes
    meta: bytes
    connectionSettings: bytes
    connectUsers: UidList
    domainUid: bytes
    jitSettings: bytes
    def __init__(self, recordUid: _Optional[bytes] = ..., networkUid: _Optional[bytes] = ..., adminUid: _Optional[bytes] = ..., meta: _Optional[bytes] = ..., connectionSettings: _Optional[bytes] = ..., connectUsers: _Optional[_Union[UidList, _Mapping]] = ..., domainUid: _Optional[bytes] = ..., jitSettings: _Optional[bytes] = ...) -> None: ...



import datetime
import logging
import os
import sqlite3
import sys
from typing import Dict, Tuple

from .. import api, crypto, utils
from ..commands.helpers.enterprise import user_has_privilege, is_addon_enabled
from ..error import CommandError, Error, KeeperApiError
from ..params import KeeperParams
from ..proto import enterprise_pb2
from . import sqlite_storage, sox_data
from .storage_types import StorageRecord, StorageUser, StorageUserRecordLink, StorageTeam, \
    StorageRecordPermissions, StorageTeamUserLink, StorageSharedFolderRecordLink, StorageSharedFolderUserLink, \
    StorageSharedFolderTeamLink

API_SOX_REQUEST_USER_LIMIT = 1000


def validate_data_access(params, cmd=''):
    privilege = 'run_compliance_reports'
    addon = 'compliance_report'
    msg_no_priv = 'You do not have the required privilege to run a Compliance Report.'
    msg_no_addon = ('Compliance reports add-on is required to perform this action. '
                    'Please contact your administrator to enable this feature.')
    error_msg = msg_no_priv if not user_has_privilege(params, privilege) \
        else msg_no_addon if not is_addon_enabled(params, addon) \
        else None
    if error_msg:
        raise CommandError(cmd, error_msg)


def is_compliance_reporting_enabled(params):
    privilege = 'run_compliance_reports'
    addon = 'compliance_report'
    return user_has_privilege(params, privilege) and is_addon_enabled(params, addon)


def encrypt_data(params, data): # type: (KeeperParams, str) -> bytes
    data_bytes = utils.string_to_bytes(data) if data else b''
    encrypted = ''
    try:
        encrypted = crypto.encrypt_aes_v1(data_bytes, params.enterprise.get('unencrypted_tree_key'))
    except Error as e:
        logging.info(f'Error encrypting data, type = {type(data)}, value = {data}, message = {e.message}')
    return encrypted


def get_sox_database_name(params, enterprise_id):  # type: (KeeperParams, int) -> str
    path = os.path.dirname(os.path.abspath(params.config_filename or '1'))
    return os.path.join(path, f'sox_{enterprise_id}.db')


def get_prelim_data(params, enterprise_id=0, rebuild=False, min_updated=0, cache_only=False, no_cache=False, shared_only=False):
    # type: (KeeperParams, int, bool, int, bool, bool, bool) -> sox_data.SoxData
    def sync_down(name_by_id, store):  # type: (Dict[int, str], sqlite_storage.SqliteSoxStorage) ->  None
        def to_storage_types(user_data, username_lookup):
            def to_record_entity(record):
                record_uid_bytes = record.recordUid
                record_uid = utils.base64_url_encode(record_uid_bytes)
                entity = store.get_records().get_entity(record_uid) or StorageRecord()
                entity.record_uid_bytes = record_uid_bytes
                entity.record_uid = record_uid
                entity.encrypted_data = record.encryptedData
                entity.shared = record.shared
                return entity

            def to_user_entity(user, email_lookup):
                entity = store.get_users().get_entity(user.enterpriseUserId) or StorageUser()
                entity.status = user.status
                user_id = user.enterpriseUserId
                entity.user_uid = user_id
                email = email_lookup.get(user_id)
                entity.email = encrypt_data(params, email) if email else b''
                return entity

            def to_user_record_link(uuid, ruid):
                link = StorageUserRecordLink()
                link.user_uid = uuid
                link.record_uid = ruid
                return link

            user_ent = to_user_entity(user_data, username_lookup)
            record_ents = [to_record_entity(x) for x in user_data.auditUserRecords if x.encryptedData]
            user_rec_links = {to_user_record_link(user_ent.user_uid, record_ent.record_uid) for record_ent in
                              record_ents}
            return user_ent, record_ents, user_rec_links

        def print_status(users_loaded, users_total, records_loaded, records_total):
            print('\r' + (100 * ' '), file=sys.stderr, end='', flush=True)
            print(f'\rLoading record information - Users: {users_loaded}/{users_total}, Current Batch: {records_loaded}/{records_total}', file=sys.stderr, end='', flush=True)

        def sync_all():
            user_ids = list(user_lookup.keys())
            users_total = len(user_ids)
            records_total = 0
            print_status(0, users_total, 0, records_total)
            users, records, links = [], [], []
            chunk_size = 1
            problem_ids = set()
            while user_ids:
                token = b''
                chunk = user_ids[:chunk_size]
                user_ids = user_ids[chunk_size:]
                rq = enterprise_pb2.PreliminaryComplianceDataRequest()
                rq.enterpriseUserIds.extend(chunk)
                rq.includeNonShared = not shared_only
                has_more = True
                current_batch_loaded = 0
                while has_more:
                    rq.continuationToken = token or rq.continuationToken
                    rq.includeTotalMatchingRecordsInFirstResponse = True
                    endpoint = 'enterprise/get_preliminary_compliance_data'
                    rs_type = enterprise_pb2.PreliminaryComplianceDataResponse
                    try:
                        rs = api.communicate_rest(params, rq, endpoint, rs_type=rs_type)
                        has_more = rs.hasMore
                        if rs.totalMatchingRecords:
                            current_batch_loaded = 0
                            records_total = rs.totalMatchingRecords
                            if records_total < 20 * API_SOX_REQUEST_USER_LIMIT:
                               # Adjust chunk size to optimize queries
                               chunk_size = min(chunk_size * 2, API_SOX_REQUEST_USER_LIMIT)
                        token = rs.continuationToken
                        for user_data in rs.auditUserData:
                            t_user, t_recs, t_links = to_storage_types(user_data, name_by_id)
                            users += [t_user]
                            records += t_recs
                            current_batch_loaded += len(t_recs)
                            print_status(users_total - len(user_ids), users_total, current_batch_loaded, records_total)
                            links += t_links
                        if not has_more:
                            print_status(users_total - len(user_ids), users_total, records_total, records_total)
                    except KeeperApiError as kae:
                        if kae.message.lower() == 'gateway_timeout':
                            # Break up the request if the number of corresponding records exceeds the backend's limit
                            if chunk_size > 1:
                                chunk_size = 1
                                user_ids = [*chunk, *user_ids]
                            else:
                                problem_ids.update(*chunk)
                            break
                        else:
                            raise kae
                    except Exception as ex:
                        raise ex
            if problem_ids:
                problem_emails = '\n'.join([name_by_id.get(id) for id in problem_ids])
                logging.error(f'Data could not fetched for the following users: \n{problem_emails}')

            store.rebuild_prelim_data(users, records, links)
        sync_all()
        print('', file=sys.stderr, flush=True)

    validate_data_access(params)
    enterprise_id = enterprise_id or next(((x['node_id'] >> 32) for x in params.enterprise['nodes']), 0)
    database_name = get_sox_database_name(params, enterprise_id)
    tree_key = params.enterprise['unencrypted_tree_key']
    ecc_key = utils.base64_url_decode(params.enterprise['keys']['ecc_encrypted_private_key'])
    ecc_key = crypto.decrypt_aes_v2(ecc_key, tree_key)
    key = crypto.load_ec_private_key(ecc_key)
    storage = sqlite_storage.SqliteSoxStorage(
        get_connection=lambda: sqlite3.connect(database_name), owner=params.user, database_name=database_name
    )
    last_updated = storage.last_prelim_data_update
    only_shared_cached = storage.shared_records_only
    refresh_data = rebuild or not last_updated or min_updated > last_updated or only_shared_cached and not shared_only
    if refresh_data and not cache_only:
        user_lookup = {x['enterprise_user_id']: x['username'] for x in params.enterprise.get('users', [])}
        storage.clear_non_aging_data()
        sync_down(user_lookup, storage)
        storage.set_shared_records_only(shared_only)
    return sox_data.SoxData(params, storage=storage, no_cache=no_cache)


def get_compliance_data(params, node_id, enterprise_id=0, rebuild=False, min_updated=0, no_cache=False, shared_only=False):
    def sync_down(sdata, node_uid, user_node_id_lookup):
        recs_processed = 0
        def print_status(pct_done):
            print('\r' + (100 * ' '), file=sys.stderr, end='', flush=True)
            print(f'\rLoading compliance data - {pct_done * 100:.2f}%', file=sys.stderr, end='', flush=True)

        def run_sync_tasks():
            def do_tasks():
                print_status(0)
                users_uids = [int(uid) for uid in sdata.get_users()]
                record_uids_raw = [rec.record_uid_bytes for rec in sdata.get_records().values()]
                max_len = API_SOX_REQUEST_USER_LIMIT
                total_ruids = len(record_uids_raw)
                ruid_chunks = [record_uids_raw[x:x + max_len] for x in range(0, total_ruids, max_len)]
                for chunk in ruid_chunks:
                    sync_chunk(chunk, users_uids)
                sdata.storage.set_compliance_data_updated()
                print('', file=sys.stderr, flush=True)

            do_tasks()

        def sync_chunk(chunk, uuids):
            rs = fetch_response(raw_ruids=chunk, user_uids=uuids)
            save_response(rs)

        def fetch_response(raw_ruids, user_uids):
            rq = enterprise_pb2.ComplianceReportRequest()
            rq.saveReport = False
            rq.reportName = f'Compliance Report on {datetime.datetime.now()}'
            report_run = rq.complianceReportRun
            report_run.users.extend(user_uids)
            report_run.records.extend(raw_ruids)
            caf = report_run.reportCriteriaAndFilter
            caf.nodeId = node_uid
            caf.criteria.includeNonShared = not shared_only
            endpoint = 'enterprise/run_compliance_report'
            return api.communicate_rest(params, rq, endpoint, rs_type=enterprise_pb2.ComplianceReportResponse)

        anon_id = 0

        def save_response(rs):
            def hash_anon_ids(response):
                # create new user uid for each anonymous user (uid >> 32 == 0)
                anon_ids = dict()
                nonlocal anon_id
                for up in rs.userProfiles:
                    user_id = up.enterpriseUserId
                    if not user_id >> 32:
                        new_id = user_id + anon_id
                        anon_ids[user_id] = new_id
                        up.enterpriseUserId = new_id
                        anon_id = new_id
                for ur in rs.userRecords:
                    user_id = ur.enterpriseUserId
                    ur.enterpriseUserId = anon_ids.get(user_id, user_id)
                for folder in rs.sharedFolderUsers:
                    for idx, user_id in enumerate(folder.enterpriseUserIds):
                        folder.enterpriseUserIds[idx] = anon_ids.get(user_id, user_id)
                return response

            save_all_types(hash_anon_ids(rs))

        def save_all_types(rs):
            save_users(rs.userProfiles)
            save_records(rs.auditRecords)
            save_teams(rs.auditTeams)
            save_shared_folders_records(rs.sharedFolderRecords)
            save_shared_folder_users(rs.sharedFolderUsers)
            save_shared_folder_teams(rs.sharedFolderTeams)
            save_record_permissions(rs.sharedFolderRecords, rs.userRecords)
            save_team_users(rs.auditTeamUsers)

        def save_users(user_profiles):
            entities = []
            for up in user_profiles:
                entity = sdata.storage.users.get_entity(up.enterpriseUserId) or StorageUser()
                entity.user_uid = entity.user_uid or up.enterpriseUserId
                entity.email = entity.email or encrypt_data(params, up.email)
                entity.job_title = entity.job_title or encrypt_data(params, up.jobTitle)
                entity.full_name = entity.full_name or encrypt_data(params, up.fullName)
                user_node = user_node_id_lookup.get(up.enterpriseUserId)
                entity.node_id = user_node
                entities.append(entity)
            sdata.storage.users.put_entities(entities)

        def save_teams(audit_teams):
            entities = []
            for team in audit_teams:
                team_uid = utils.base64_url_encode(team.teamUid)
                entity = sdata.storage.teams.get_entity(team_uid) or StorageTeam()
                entity.team_uid = team_uid
                entity.team_name = team.teamName
                entity.restrict_edit = team.restrictEdit
                entity.restrict_share = team.restrictShare
                entities.append(entity)
            sdata.storage.teams.put_entities(entities)

        def save_team_users(audit_team_users):
            links = []
            for team in audit_team_users:
                team_uid = utils.base64_url_encode(team.teamUid)
                links.extend([StorageTeamUserLink(team_uid, user_uid) for user_uid in team.enterpriseUserIds])
            sdata.storage.get_team_user_links().put_links(links)

        def save_shared_folders_records(sf_records):
            links = []
            for folder in sf_records:
                folder_uid = utils.base64_url_encode(folder.sharedFolderUid)
                for rp in folder.recordPermissions:
                    record_uid = utils.base64_url_encode(rp.recordUid)
                    link = StorageSharedFolderRecordLink(folder_uid, record_uid, rp.permissionBits)
                    links.append(link)
            sdata.storage.get_sf_record_links().put_links(links)

        def save_record_permissions(sf_records, user_records):
            def update_permissions_lookup(perm_lookup, record_uid, enterprise_user_id, perm_bits):
                lookup_key = record_uid, enterprise_user_id
                p_bits = perm_lookup.get(lookup_key, perm_bits)
                perm_lookup.update({lookup_key: perm_bits | p_bits})

            # Aggregate record-permissions for each user-record pair
            rec_perms_lookup = dict()   # type: Dict[Tuple[str, int], int]
            # Save share-admin permissions
            for folder in sf_records:
                rec_perms = folder.recordPermissions
                for sar in folder.shareAdminRecords:
                    rec_perm_idxs = sar.recordPermissionIndexes
                    sar_perms = [rp for idx, rp in enumerate(rec_perms) if idx in rec_perm_idxs]
                    rec_uids = [utils.base64_url_encode(rp.recordUid) for rp in sar_perms]
                    for ruid in rec_uids:
                        update_permissions_lookup(rec_perms_lookup, ruid, sar.enterpriseUserId, 16)

            for ur in user_records:
                for rp in ur.recordPermissions:
                    ruid = utils.base64_url_encode(rp.recordUid)
                    update_permissions_lookup(rec_perms_lookup, ruid, ur.enterpriseUserId, rp.permissionBits)

            links = []
            for k, v in rec_perms_lookup.items():
                rec_uid, user_id = k
                link = StorageRecordPermissions(rec_uid, user_id, v)
                links.append(link)
            sdata.storage.get_record_permissions().put_links(links)

        def save_shared_folder_users(sf_users):
            links = []
            for folder in sf_users:
                folder_uid = utils.base64_url_encode(folder.sharedFolderUid)
                links.extend([StorageSharedFolderUserLink(folder_uid, uuid) for uuid in folder.enterpriseUserIds])
            sdata.storage.get_sf_user_links().put_links(links)

        def save_shared_folder_teams(sf_teams):
            links = []
            for folder in sf_teams:
                folder_uid = utils.base64_url_encode(folder.sharedFolderUid)
                for tuid in folder.teamUids:
                    team_uid = utils.base64_url_encode(tuid)
                    links.append(StorageSharedFolderTeamLink(folder_uid, team_uid))
            sdata.storage.get_sf_team_links().put_links(links)

        def save_records(records):
            entities = []
            for record in records:
                nonlocal recs_processed
                recs_processed += 1
                print_status(recs_processed/len(sdata.get_records()))
                rec_uid = utils.base64_url_encode(record.recordUid)
                entity = sdata.storage.records.get_entity(rec_uid)
                if entity:
                    entity.in_trash = record.inTrash
                    entity.has_attachments = record.hasAttachments
                    entities.append(entity)
            sdata.storage.records.put_entities(entities)

        run_sync_tasks()

    sd = get_prelim_data(params, enterprise_id, rebuild=rebuild, min_updated=min_updated, cache_only=not min_updated, shared_only=shared_only)
    last_compliance_data_update = sd.storage.last_compliance_data_update
    refresh_data = rebuild or min_updated > last_compliance_data_update
    if refresh_data:
        enterprise_users = params.enterprise.get('users', [])
        user_node_ids = {e_user.get('enterprise_user_id'): e_user.get('node_id') for e_user in enterprise_users}
        sync_down(sd, node_id, user_node_id_lookup=user_node_ids)
    rebuild_task = sox_data.RebuildTask(is_full_sync=False, load_compliance_data=True)
    sd.rebuild_data(rebuild_task, no_cache=no_cache)
    return sd


def get_node_id(params, name):
    name = int(name) if isinstance(name, str) and name.isdecimal() else name
    nodes = params.enterprise['nodes']
    root_node_id = nodes[0].get('node_id', 0)
    node_ids = (n.get('node_id') for n in nodes)
    node_id_lookup = {n.get('data').get('displayname'): n.get('node_id') for n in nodes}
    node_id = node_id_lookup.get(name) if name in node_id_lookup \
        else name if name in node_ids \
        else root_node_id
    return node_id
